# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import os  # load the python os default module
import sys  # laod the python sys default module
import copy
import warnings

import numpy as np
import pandas as pd

import pandapower as pp
import scipy

from grid2op.dtypes import dt_int, dt_float, dt_bool
from grid2op.Backend.backend import Backend
from grid2op.Exceptions import *


class EducPandaPowerBackend(Backend):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        This class does not work.

        It is mainly presented for educational purpose as example on coding your own backend.

        It is derived from PandaPowerBackend, but without all the "optimization" that could make the
        resulting backend harder to read.

    This module presents an example of an implementation of a `grid2op.Backend` when using the powerflow
    implementation "pandapower" available at `PandaPower <https://www.pandapower.org/>`_ for more details about
    this backend. This file is provided as an example of a proper :class:`grid2op.Backend.Backend` implementation.

    This backend currently does not work with 3 winding transformers and other exotic object.

    As explained in the `grid2op.Backend` module, every module must inherit the `grid2op.Backend` class.

    We illustrate here how to set up a backend with what we think is "rather standard" in the powersystem
    eco system.

    Please consult the documentation at :ref:`create-backend-module` for more information.

    You have at your disposal:

    - a tool that is able to compute power flows from a given grid in a given format (in this case call
      `pandapower.runpf(pandapower_grid)`)
    - a tool that is able to load a powergrid from a file store on the hard drive.

    We try to find a good compromise between the size of the code (for clarity) and the "closeness to a working
    code".

    For a complete working example, relatively optimized (but much less readable) please have a look at the
    real :class:`grid2op.Backend.PandaPowerBackend` class.

    """

    def __init__(self,
                 detailed_infos_for_cascading_failures=False,
                 can_be_copied=True):
        """
        Nothing much to do here except initializing what you would need (a tensorflow session, link to some
        external dependencies etc.)

        Nothing much for this example.

        Parameters
        ----------
        detailed_infos_for_cascading_failures: ``bool``
            See the documentation of :class:`grid2op.Backend.Backend.__init__` for more information
        """
        Backend.__init__(
            self,
            detailed_infos_for_cascading_failures=detailed_infos_for_cascading_failures,
            can_be_copied=can_be_copied,
            # extra arguments that might be needed for building such a backend 
            # these extra kwargs will be stored (without copy) in the
            # base class and used when another backend will be created
            # for example in the Runner class.
        )
        warnings.warn(
            "This backend is used for demonstration purpose only, you should not use it under any "
            "circumstances. Please use grid2op.Backend.PandaPowerBackend instead"
        )
        self._nb_real_line_pandapower = None

        # NB: this instance of backend is here for academic purpose only. For clarity, it does not handle
        # neither shunt nor storage unit.

    ####### load the grid
    def load_grid(self, path=None, filename=None):
        """
        Demonstration on how you can load a powergrid and then initialize the proper grid2op attributes.

        The only thing we have to do here is to "order" the objects in each substation. Note that we don't even do it
        implicitly here (relying on default grid2op ordering).

        The only decision we had to make was concerning "grid2op powerlines" which represents both
        "pandapower transformers"
        and "pandapower powerlines".

        We decided that:

        - powerline are "before" trafo (so in the grid2op line id I will put first all powerlines, then all trafo)
        - grid2op "origin" side will be "from" side for pandapower powerline and "hv" side for pandapower trafo
        - grid2op "extremity" side will be "to" side for pandapower powerline and "lv" side for pandapower trafo


        .. note:: We use one "trick" here. Pandapower grid (as it will be the case for most format) will have one "bus"
            per substation. For grid2op, we want at least 2 busbar per substation. So we simply copy and paste the grid.

            And we will deactivate the busbar that are not connected (no element connected to it).

            This "coding" allows for easily mapping the bus id (each bus is represented with an id in pandapower)
            and whether its busbar 1 or busbar 2 (grid2op side). More precisely: busbar 1 of substation with
            id `sub_id` will have id `sub_id` and busbar 2 of the same substation will have id `sub_id + n_sub`
            (recall that n_sub is the number of substation on the grid).

            This "coding" is not optimal in the ram it takes. But we recommend you to adopt a similar one. It's
            pretty easy to change the topology using this trick, much easier than if you rely on "switches" for
            example. (But of course  you can still use switches if you really want to)

        """

        # first, handles different kind of path:
        if path is None and filename is None:
            raise RuntimeError(
                "You must provide at least one of path or file to load a powergrid."
            )
        if path is None:
            full_path = filename
        elif filename is None:
            full_path = path
        else:
            full_path = os.path.join(path, filename)
        if not os.path.exists(full_path):
            raise RuntimeError('There is no powergrid at "{}"'.format(full_path))

        # then load the grid located at the full path and store it in `self._grid`
        # raise an exception if it can't be loaded
        try:
            with warnings.catch_warnings():
                # remove deprecationg warnings for old version of pandapower
                warnings.filterwarnings("ignore", category=DeprecationWarning)
                self._grid = pp.from_json(full_path)
        except Exception as exc_:
            raise BackendError(
                f'Impossible to load the powergrid located at "{full_path}". Please '
                f"check the file exist and that the file represent a valid pandapower "
                f"grid. For your information, the error is:\n{exc_}"
            )

        ######################################################################
        # this part is due to the "modeling" of the topology FOR THIS EXAMPLE
        # remember (see docstring of this function) that we "duplicate the buses" to code more easily the
        # topology modification (instead of relying on the `switches`)

        # first we remember the number of substation
        self.n_sub = self._grid.bus.shape[0]
        # and then we duplicate the bus
        add_topo = copy.deepcopy(self._grid.bus)
        add_topo.index += add_topo.shape[0]
        add_topo["in_service"] = False
        self._grid.bus = pd.concat((self._grid.bus, add_topo))
        self._nb_real_line_pandapower = self._grid.line.shape[0]
        # i do a powerflow to initialize the "results" dataframes
        # this last step is internal to pandapower
        pp.runpp(self._grid, check_connectivity=False)
        ######################################################################

        # and now we initialize the number of each of the elements
        self.n_line = (
            self._grid.line.shape[0] + self._grid.trafo.shape[0]
        )  # trafo are powerline for grid2op !
        self.n_gen = self._grid.gen.shape[0]
        self.n_load = self._grid.load.shape[0]
        # self.n_sub  # already initialize above

        # initialize the number of elements per substation
        # now export to grid2op the substation to which objects are connected
        self.load_to_subid = copy.deepcopy(self._grid.load["bus"])
        self.gen_to_subid = copy.deepcopy(self._grid.gen["bus"])
        # here we just decide (but that is a convention we could have done it differently)
        # that "origin side" (grid2op) corresponds to "from_bus" from pandapower line and "hv_bus" for
        # pandapower trafo.
        self.line_or_to_subid = np.concatenate(
            (
                copy.deepcopy(self._grid.line["from_bus"]),
                copy.deepcopy(self._grid.trafo["hv_bus"]),
            )
        )
        self.line_ex_to_subid = np.concatenate(
            (
                copy.deepcopy(self._grid.line["to_bus"]),
                copy.deepcopy(self._grid.trafo["lv_bus"]),
            )
        )

        # and now we don't forget to initialize the rest
        self._compute_pos_big_topo()  # we highly recommend you to call this !

        # and now the thermal limit
        self.thermal_limit_a = 1000. * np.concatenate(
            (
                self._grid.line["max_i_ka"].values,
                self._grid.trafo["sn_mva"].values
                / (np.sqrt(3) * self._grid.trafo["vn_hv_kv"].values),
            )
        )
        self.thermal_limit_a = self.thermal_limit_a.astype(dt_float)

        # NB: this instance of backend is here for academic purpose only. For clarity, it does not handle
        # neither shunt nor storage unit.
        type(self).shunts_data_available = False
        type(self).set_no_storage()

    ###### modify the grid
    def apply_action(self, backendAction=None):
        """
        Here the implementation of the "modify the grid" function.

        From the documentation, it's pretty straightforward, even though the implementation takes ~70 lines of code.
        Most of them being direct copy paste from the examples in the documentation.
        """
        if backendAction is None:
            return

        (
            active_bus,
            (prod_p, prod_v, load_p, load_q, storage),
            _,
            shunts__,
        ) = backendAction()

        for gen_id, new_p in prod_p:
            self._grid.gen["p_mw"].iloc[gen_id] = new_p
        for gen_id, new_v in prod_v:
            self._grid.gen["vm_pu"].iloc[gen_id] = new_v  # but new_v is not pu !
            self._grid.gen["vm_pu"].iloc[gen_id] /= self._grid.bus["vn_kv"][
                self.gen_to_subid[gen_id]
            ]  # now it is :-)

        for load_id, new_p in load_p:
            self._grid.load["p_mw"].iloc[load_id] = new_p
        for load_id, new_q in load_q:
            self._grid.load["q_mvar"].iloc[load_id] = new_q

        # now i deal with the topology
        loads_bus = backendAction.get_loads_bus()
        for load_id, new_bus in loads_bus:
            if new_bus == -1:
                self._grid.load["in_service"][load_id] = False
            else:
                self._grid.load["in_service"][load_id] = True
                # this formula is really convenient because we decided to duplicated buses in each substation.
                # and decided that: bus 1 of a substation with id `sub_id` will have id `sub_id` and
                # bus 2 of the same substation will have id `sub_id + n_substation`
                self._grid.load["bus"][load_id] = (
                    self.load_to_subid[load_id] + (new_bus - 1) * self.n_sub
                )

        gens_bus = backendAction.get_gens_bus()
        for gen_id, new_bus in gens_bus:
            if new_bus == -1:
                self._grid.gen["in_service"][gen_id] = False
            else:
                self._grid.gen["in_service"][gen_id] = True
                self._grid.gen["bus"][gen_id] = (
                    self.gen_to_subid[gen_id] + (new_bus - 1) * self.n_sub
                )

        lines_or_bus = backendAction.get_lines_or_bus()
        for line_id, new_bus in lines_or_bus:
            if line_id < self._nb_real_line_pandapower:
                dt = self._grid.line
                key = "from_bus"
                line_id_db = line_id
            else:
                dt = self._grid.trafo
                key = "hv_bus"
                line_id_db = line_id - self._nb_real_line_pandapower

            if new_bus == -1:
                dt["in_service"][line_id_db] = False
            else:
                dt["in_service"][line_id_db] = True
                dt[key][line_id_db] = (
                    self.line_or_to_subid[line_id] + (new_bus - 1) * self.n_sub
                )

        lines_ex_bus = backendAction.get_lines_ex_bus()
        for line_id, new_bus in lines_ex_bus:
            if line_id < self._nb_real_line_pandapower:
                dt = self._grid.line
                key = "to_bus"
                line_id_db = line_id
            else:
                dt = self._grid.trafo
                key = "lv_bus"
                line_id_db = line_id - self._nb_real_line_pandapower

            if new_bus == -1:
                dt["in_service"][line_id_db] = False
            else:
                dt["in_service"][line_id_db] = True
                dt[key][line_id_db] = (
                    self.line_ex_to_subid[line_id] + (new_bus - 1) * self.n_sub
                )

        # spec
        bus_is = self._grid.bus["in_service"]
        for i, (bus1_status, bus2_status) in enumerate(active_bus):
            bus_is[i] = bus1_status  # no iloc for bus, don't ask me why please :-/
            bus_is[i + self.n_sub] = bus2_status

    ###### computes powerflow
    def runpf(self, is_dc=False):
        """
        Now we just perform a powerflow with pandapower which can be done with either `rundcpp` for dc powerflow
        or with `runpp` for AC powerflow.

        This is really a 2 lines code. It is a bit more verbose here because we took care of silencing some
        warnings and try / catch some possible exceptions.

        """
        try:
            with warnings.catch_warnings():
                # remove the warning if _grid non connex. And it that case load flow as not converged
                warnings.filterwarnings(
                    "ignore", category=scipy.sparse.linalg.MatrixRankWarning
                )
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                if is_dc:
                    pp.rundcpp(self._grid, check_connectivity=False)
                else:
                    pp.runpp(self._grid, check_connectivity=False)
                return self._grid.converged, None
        except pp.powerflow.LoadflowNotConverged as exc_:
            # of the powerflow has not converged, results are Nan
            return False, exc_

    ###### getters
    def get_topo_vect(self):
        """
        Retrieve the bus to which the objects are connected based on the information stored on the grid.

        This is fairly simple, again, because we choose to explicitly represents 2 buses per substation.

        Function is verbose (~40 lines of code), but pretty straightforward.
        """
        res = np.full(self.dim_topo, fill_value=np.NaN, dtype=dt_int)

        line_status = self.get_line_status()

        i = 0
        for row in self._grid.line[["from_bus", "to_bus"]].values:
            bus_or_id = row[0]
            bus_ex_id = row[1]
            if line_status[i]:
                res[self.line_or_pos_topo_vect[i]] = (
                    1 if bus_or_id == self.line_or_to_subid[i] else 2
                )
                res[self.line_ex_pos_topo_vect[i]] = (
                    1 if bus_ex_id == self.line_ex_to_subid[i] else 2
                )
            else:
                res[self.line_or_pos_topo_vect[i]] = -1
                res[self.line_ex_pos_topo_vect[i]] = -1
            i += 1

        nb = self._nb_real_line_pandapower
        i = 0
        for row in self._grid.trafo[["hv_bus", "lv_bus"]].values:
            bus_or_id = row[0]
            bus_ex_id = row[1]

            j = i + nb
            if line_status[j]:
                res[self.line_or_pos_topo_vect[j]] = (
                    1 if bus_or_id == self.line_or_to_subid[j] else 2
                )
                res[self.line_ex_pos_topo_vect[j]] = (
                    1 if bus_ex_id == self.line_ex_to_subid[j] else 2
                )
            else:
                res[self.line_or_pos_topo_vect[j]] = -1
                res[self.line_ex_pos_topo_vect[j]] = -1
            i += 1

        i = 0
        for bus_id in self._grid.gen["bus"].values:
            res[self.gen_pos_topo_vect[i]] = 1 if bus_id == self.gen_to_subid[i] else 2
            i += 1

        i = 0
        for bus_id in self._grid.load["bus"].values:
            res[self.load_pos_topo_vect[i]] = (
                1 if bus_id == self.load_to_subid[i] else 2
            )
            i += 1
        return res

    def generators_info(self):
        """
        We chose to keep the same order in grid2op and in pandapower. So we just return the correct values.
        """
        # carefull with copy / deep copy
        prod_p = self._grid.res_gen["p_mw"].values.astype(dt_float)
        prod_q = self._grid.res_gen["q_mvar"].values.astype(dt_float)
        prod_v = self._grid.res_gen["vm_pu"].values.astype(dt_float)  # in pu
        prod_v *= (
            self._grid.bus["vn_kv"].iloc[self.gen_to_subid].values.astype(dt_float)
        )  # in kV
        return prod_p, prod_q, prod_v

    def loads_info(self):
        """
        We chose to keep the same order in grid2op and in pandapower. So we just return the correct values.
        """
        # carefull with copy / deep copy
        load_p = self._grid.res_load["p_mw"].values.astype(dt_float)
        load_q = self._grid.res_load["q_mvar"].values.astype(dt_float)
        load_v = self._grid.res_bus.loc[self._grid.load["bus"].values][
            "vm_pu"
        ].values.astype(
            dt_float
        )  # in pu
        load_v *= self._grid.bus.loc[self._grid.load["bus"].values][
            "vn_kv"
        ].values.astype(
            dt_float
        )  # in kV
        return load_p, load_q, load_v

    def _aux_get_line_info(self, colname_powerline, colname_trafo):
        """
        concatenate the information of powerlines and trafo using the convention that "powerlines go first"
        """
        res = np.concatenate(
            (
                self._grid.res_line[colname_powerline].values,
                self._grid.res_trafo[colname_trafo].values,
            )
        )
        return res

    def lines_or_info(self):
        """
        Main method to retrieve the information at the "origin" side of the powerlines and transformers.

        We simply need to follow the convention we adopted:

        - origin side (grid2op) will be "from" side for pandapower powerline
        - origin side (grid2op) will be "hv" side for pandapower trafo
        - we chose to first have powerlines, then transformers

        (convention chosen in :func:`EducPandaPowerBackend.load_grid`)

        """
        p_or = self._aux_get_line_info("p_from_mw", "p_hv_mw")
        q_or = self._aux_get_line_info("q_from_mvar", "q_hv_mvar")
        v_or = self._aux_get_line_info("vm_from_pu", "vm_hv_pu")
        a_or = self._aux_get_line_info("i_from_ka", "i_hv_ka") * 1000
        return p_or, q_or, v_or, a_or

    def lines_ex_info(self):
        """
        Main method to retrieve the information at the "extremity" side of the powerlines and transformers.

        We simply need to follow the convention we adopted:

        - extremity side (grid2op) will be "to" side for pandapower powerline
        - extremity side (grid2op) will be "lv" side for pandapower trafo
        - we chose to first have powerlines, then transformers

        (convention chosen in function :func:`EducPandaPowerBackend.load_grid`)

        """
        p_ex = self._aux_get_line_info("p_to_mw", "p_lv_mw")
        q_ex = self._aux_get_line_info("q_to_mvar", "q_lv_mvar")
        v_ex = self._aux_get_line_info("vm_to_pu", "vm_lv_pu")
        a_ex = self._aux_get_line_info("i_to_ka", "i_lv_ka") * 1000
        return p_ex, q_ex, v_ex, a_ex

    # other less important method that you will need to implement
    def get_line_status(self):
        """
        you might consider implementing it

        .. warning::  /!\\\\ This is a not a "main method" but you might want to implement
            it for a new backend (default implementation most likely not efficient at all). /!\\\\
        """
        return np.concatenate(
            (
                self._grid.line["in_service"].values,
                self._grid.trafo["in_service"].values,
            )
        ).astype(dt_bool)

    def _disconnect_line(self, id_):
        """
        you might consider implementing it

        .. warning:: /!\\\\ This is a not a "main method" but you might want to implement
            it for a new backend (default implementation most likely not efficient at all). /!\\\\
        """
        if id_ < self._nb_real_line_pandapower:
            self._grid.line["in_service"].iloc[id_] = False
        else:
            self._grid.trafo["in_service"].iloc[
                id_ - self._nb_real_line_pandapower
            ] = False

    def copy(self):
        """
        you might consider implementing it

        .. warning:: /!\\\\ This is a not a "main method" but you might want to implement
            it for a new backend (default implementation most likely not efficient at all). /!\\\\

        Nothing really crazy here

        Performs a deep copy of the power :attr:`_grid`.
        As pandapower is pure python, the deep copy operator is perfectly suited for the task.
        """
        res = copy.deepcopy(self)
        return res

    def reset(self, path=None, grid_filename=None):
        """
        you might consider implementing it

        .. warning:: /!\\\\ This is a not a "main method" but you might want to implement
            it for a new backend (default implementation most likely not efficient at all). /!\\\\

        Reset the grid to the original state
        """

        # set everything to its proper bus (this is because we used a specific way to represent
        # the topology for this example by chosing not to use swtiches, but to double the number of
        # buses per "substation"
        self._grid.line["from_bus"].iloc[:] = self.line_or_to_subid[
            : self._nb_real_line_pandapower
        ]
        self._grid.trafo["hv_bus"].iloc[:] = self.line_or_to_subid[
            self._nb_real_line_pandapower :
        ]
        self._grid.line["to_bus"].iloc[:] = self.line_ex_to_subid[
            : self._nb_real_line_pandapower
        ]
        self._grid.trafo["lv_bus"].iloc[:] = self.line_ex_to_subid[
            self._nb_real_line_pandapower :
        ]
        self._grid.load["bus"].iloc[:] = self.load_to_subid
        self._grid.gen["bus"].iloc[:] = self.gen_to_subid

        # originally everything is in service
        self._grid.line["in_service"].iloc[:] = True
        self._grid.trafo["in_service"].iloc[:] = True
        self._grid.load["in_service"].iloc[:] = True
        self._grid.gen["in_service"].iloc[:] = True

        self._grid.bus["in_service"].iloc[: self.n_sub] = True
        self._grid.bus["in_service"].iloc[self.n_sub :] = False

    def close(self):
        """
        you might consider implementing it

        .. warning:: /!\\\\ This is a not a "main method" but you might want to implement
            it for a new backend (default implementation most likely not efficient at all). /!\\\\

        Called when the :class:`grid2op;Environment` has terminated, this function only reset the grid to a state
        where it has not been loaded.
        """
        del self._grid
        self._grid = None



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import os
import typing as t
import warnings

from flask import Flask
from flask import make_response, jsonify
from flask import request
from collections.abc import Iterable

from grid2op.rest_server.env_cache import EnvCache
import argparse

try:
    import ujson
    from flask.json import JSONEncoder, JSONDecoder

    # define the encoder
    class CustomJSONEncoder(JSONEncoder):
        def default(self, obj):
            try:
                return ujson.dumps(obj)
            except TypeError:
                return JSONEncoder.default(self, obj)

    # define the decoder
    class CustomJSONDecoder(JSONDecoder):
        def dump(
            self,
            obj: t.Any,
            fp: t.IO[str],
            app: t.Optional["Flask"] = None,
            **kwargs: t.Any,
        ) -> None:
            try:
                return ujson.dump(obj=obj, fp=fp)
            except TypeError:
                return CustomJSONDecoder.dump(self, obj, fp, app, **kwargs)

        def dumps(
            self, obj: t.Any, app: t.Optional["Flask"] = None, **kwargs: t.Any
        ) -> str:
            try:
                return ujson.dumps(obj=obj)
            except TypeError:
                return CustomJSONDecoder.dumps(self, obj, app, **kwargs)

        def loads(
            self, s: str, app: t.Optional["Flask"] = None, **kwargs: t.Any
        ) -> t.Any:
            try:
                return ujson.loads(s)
            except TypeError:
                return CustomJSONDecoder.loads(self, s, app, **kwargs)

        def load(
            self, fp: t.IO[str], app: t.Optional["Flask"] = None, **kwargs: t.Any
        ) -> t.Any:
            try:
                return ujson.load(fp=fp)
            except TypeError:
                return CustomJSONDecoder.load(self, fp, app, **kwargs)

    UJSON_AS_JSON = True
except ImportError as exc:
    warnings.warn("ujson not available, expect some degraded performance")
    UJSON_AS_JSON = False

ENV_CACHE = EnvCache(UJSON_AS_JSON)

app = Flask(__name__)
if UJSON_AS_JSON:
    app.json_encoder = CustomJSONEncoder
    app.json_decoder = CustomJSONDecoder

# TODO for improved security, not sure it's needed
if False:
    from flask_wtf.csrf import CSRFProtect

    csrf = CSRFProtect()
    csrf.init_app(app)
    # set the env variable this way before starting : `set WTF_CSRF_SECRET_KEY=...`
    SECRET_KEY = os.urandom(32)
    app.config["SECRET_KEY"] = SECRET_KEY


@app.route("/")
def index():
    return (
        "Welcome to grid2op. This small server lets you use grid2op as an web service to use some grid2op "
        "features for example in different computer languages. See the documentation for more information."
        "(work in progress)"
        "(alpha mode at the moment)"
    )


@app.route("/make/<env_name>")
def make_env(env_name):
    """
    This function lets you create an environment with the name "env_name".

    It is equivalent to perform a call to `grid2op.make(env_name)` followed by `obs = env.reset()`

    TODO support parameters and backend and all the other kwargs of make

    Notes
    ------
    This is a simple `get` request.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created
    - "obs" : the json representation of the observation you get after creation of this environment

    """
    id_, obs, exc_ = ENV_CACHE.insert_env(env_name)
    if exc_ is not None:
        return make_response(
            jsonify(
                {
                    "error": f'Impossible to create environment "{env_name}" with error:\n'
                    f"{exc_}"
                }
            ),
            400,
        )
    resp = {"id": id_, "env_name": env_name, "obs": obs}
    return jsonify(resp)


@app.route("/reset/<env_name>/<env_id>")
def reset(env_name, env_id):
    """
    This call is equivalent to do: `env.reset()` when ``env`` is the environment with id "env_id" and name "env_name"

    Notes
    ------
    This is a simple `get` request.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created
    - "obs" : the json representation of the observation you get after having reset this environment.

    """
    obs, (error_code, error_msg) = ENV_CACHE.reset(env_name, env_id)
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {"id": env_id, "env_name": env_name, "obs": obs}
    return make_response(jsonify(resp))


@app.route("/close/<env_name>/<env_id>")
def close(env_name, env_id):
    """
    This call is equivalent to do: `env.close()` when ``env`` is the environment with id "env_id" and name "env_name".

    Note that after being closed, any use of the environment will raise an error.

    Notes
    ------
    This is a simple `get` request.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created

    """
    error_code, error_msg = ENV_CACHE.close(env_name, env_id)
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {"id": env_id, "env_name": env_name}
    return make_response(jsonify(resp))


@app.route("/get_path_env/<env_name>/<env_id>")
def get_path_env(env_name, env_id):
    """
    This call is equivalent to do: `env.get_path_env()` when ``env`` is the environment with id "env_id" and name
    "env_name".

    It returns the (local to the server, not the client!) path where the environment is located.

    Notes
    ------
    This is a simple `get` request.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created
    - "path": the path of the environment (**NB** this path is local to the server!)

    """
    path, (error_code, error_msg) = ENV_CACHE.get_path_env(env_name, env_id)
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {"id": env_id, "env_name": env_name, "path": path}
    return make_response(jsonify(resp))


@app.route("/get_thermal_limit/<env_name>/<env_id>")
def get_thermal_limit(env_name, env_id):
    """
    This call is equivalent to do: `env.get_thermal_limit()` when ``env`` is the environment with id "env_id" and name
    "env_name".

    Notes
    ------
    This is a simple `get` request.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created
    - "thermal_limit": ``list`` the thermal limit for each powerline.

    """
    th_lim, (error_code, error_msg) = ENV_CACHE.get_thermal_limit(env_name, env_id)
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {"id": env_id, "env_name": env_name, "thermal_limit": th_lim}
    return make_response(jsonify(resp))


@app.route("/step/<env_name>/<env_id>", methods=["POST"])
def step(env_name, env_id):
    """
    This call is equivalent to do: `env.step(action)` when ``env`` is the environment with id "env_id" and name
    "env_name".

    Notes
    ------
    This is a `post` request.

    The payload (data) should be a json with key "action" (``dict``) representing a valid grid2op action

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created
    - "obs": the json representation of the observation you get after this step
    - "reward": the reward you get after this step (``float``)
    - "done": a flag indicating whether or not the environment has terminated (``bool``). If this flag is
      ``True`` then you need to call `reset` on this same environment (same name, same id)
      if you want to continue to use it.
    - "info": a list of detailed information returned by step (more information in the documentation of
      :func:`grid2op.Environment.BaseEnv.step`)

    """
    # handle the action part
    if not request.json or "action" not in request.json:
        make_response(
            jsonify(
                {"error": f'You need to provide an action in order to do a "step".'}
            ),
            400,
        )

    (obs, reward, done, info), (error_code, error_msg) = ENV_CACHE.step(
        env_name, env_id, request.json["action"]
    )
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {
        "id": env_id,
        "env_name": env_name,
        "obs": obs,
        "reward": reward,
        "done": done,
        "info": info,
    }
    return make_response(jsonify(resp))


@app.route("/seed/<env_name>/<env_id>", methods=["POST"])
def seed(env_name, env_id):
    """
    This call is equivalent to do: `env.seed(seed)` when ``env`` is the environment with id "env_id" and name
    "env_name".

    Notes
    ------
    This is a `post` request.

    The payload (data) should be a json with key "seed" (``int``) representing the seed (an integer) you want to use.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created
    - "seeds": the seeds used to ensure reproducibility of all the environment components
      (more information in the documentation of :func:`grid2op.Environment.BaseEnv.seed`)
    - "info": a generic text to make sure you know that you need to call reset before it has any effect.

    """
    if not request.json or "seed" not in request.json:
        make_response(
            jsonify(
                {
                    "error": f'You need to provide an action in order to "seed" the environment.'
                }
            ),
            400,
        )

    seeds, (error_code, error_msg) = ENV_CACHE.seed(
        env_name, env_id, request.json["seed"]
    )
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {
        "id": env_id,
        "env_name": env_name,
        "seeds": seeds,
        "info": "this has no effect until reset() is called",
    }
    return make_response(jsonify(resp))


@app.route("/set_id/<env_name>/<env_id>", methods=["POST"])
def set_id(env_name, env_id):
    """
    This call is equivalent to do: `env.set_id(id)` when ``env`` is the environment with id "env_id" and name
    "env_name".

    It has no effect unless "reset" is used.

    Notes
    ------
    This is a `post` request.

    The payload (data) should be a json with key "id" (``int``) representing the chronic id you want to go to.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created
    - "info": a generic text to make sure you know that you need to call reset before it has any effect.

    """
    if not request.json or "id" not in request.json:
        make_response(
            jsonify(
                {
                    "error": f'You need to provide an id in order to "set_id" the environment.'
                }
            ),
            400,
        )
    error_code, error_msg = ENV_CACHE.set_id(env_name, env_id, request.json["id"])
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {
        "id": env_id,
        "env_name": env_name,
        "info": "this has no effect until reset() is called",
    }
    return make_response(jsonify(resp))


@app.route("/set_thermal_limit/<env_name>/<env_id>", methods=["POST"])
def set_thermal_limit(env_name, env_id):
    """
    This call is equivalent to do: `env.set_thermal_limit(thermal_limits)` when
    ``env`` is the environment with id "env_id" and name "env_name".

    Notes
    ------
    This is a `post` request.

    The payload (data) should be a json with key "thermal_limits" (``list``) representing the new thermal limit you
    want to use.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created

    """
    if not request.json or "thermal_limits" not in request.json:
        make_response(
            jsonify(
                {
                    "error": f'You need to provide thermal limits in order to "set_thermal_limit" '
                    f"the environment."
                }
            ),
            400,
        )
    error_code, error_msg = ENV_CACHE.set_thermal_limit(
        env_name, env_id, request.json["thermal_limits"]
    )
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {"id": env_id, "env_name": env_name}
    return make_response(jsonify(resp))


@app.route("/fast_forward_chronics/<env_name>/<env_id>", methods=["POST"])
def fast_forward_chronics(env_name, env_id):
    """
    This call is equivalent to do: `env.fast_forward_chronics(nb_step)` when
    ``env`` is the environment with id "env_id" and name "env_name".

    Notes
    ------
    This is a `post` request.

    The payload (data) should be a json with key "nb_step" (``int``) representing the number of step you want to
    "fast forward" to.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created

    """
    if not request.json or "nb_step" not in request.json:
        make_response(
            jsonify(
                {
                    "error": f'You need to provide a number of step in order to "fast_forward_chronics" '
                    f"the environment."
                }
            ),
            400,
        )
    error_code, error_msg = ENV_CACHE.fast_forward_chronics(
        env_name, env_id, request.json["nb_step"]
    )
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {"id": env_id, "env_name": env_name}
    return make_response(jsonify(resp))


@app.route("/train_val_split/<env_name>/<env_id>", methods=["POST"])
def train_val_split(env_name, env_id):
    """
    This call is equivalent to do: `env.train_val_split(chron_id_val)` when
    ``env`` is the environment with id "env_id" and name "env_name".

    Notes
    ------
    This is a `post` request.

    The payload (data) should be a json with key "chron_id_val" (``list``)
    representing the ids of the chronics that will be put aside in the validation set.

    Returns
    -------
    A json with keys:

    - "id": the id of the environment created
    - "env_name": the name of the environment created
    - "nm_train": name of the environment you can use as training environment, that will contain all the initial
      chronics except the one specified in `chron_id_val`. You may initialize it with `make/nm_train`
    - "nm_val": name of the environment you can use as validation environment, that will contain only the chronics
      ids specified in `chron_id_val`. You may initialize it with `make/nm_val`

    """
    if not request.json or "chron_id_val" not in request.json:
        make_response(
            jsonify(
                {
                    "error": f"You need to provide with the id of the chronics that will go to "
                    f"the validation set "
                }
            ),
            400,
        )
    chron_id_val = request.json["chron_id_val"]
    if not isinstance(chron_id_val, Iterable):
        make_response(
            jsonify(
                {
                    "error": f'"chron_id_val"  should be an iterable representing the name of the '
                    f"scenarios "
                    f"you want to place in the validation set."
                }
            ),
            400,
        )
    (nm_train, nm_val), (error_code, error_msg) = ENV_CACHE.train_val_split(
        env_name, env_id, chron_id_val
    )
    if error_code is not None:
        return make_response(
            jsonify({"error": error_msg, "error_code": error_code}), 400
        )

    resp = {"id": env_id, "env_name": env_name, "nm_train": nm_train, "nm_val": nm_val}
    return make_response(jsonify(resp))


# TODO
# set_id
# set_thermal_limit
# get_thermal_limit
# fast_forward_chronics
# get_path_env
# close
# train_val_split  # not tested

# TODO
# asynch here!


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Start a mono thread / mono process grid2op environment server"
    )
    parser.add_argument(
        "--port",
        type=int,
        default=3000,
        help="On which port to start the server (default 3000)",
    )
    parser.add_argument(
        "--debug",
        type=str2bool,
        nargs="?",
        const=True,
        default=False,
        help="Start the flask server in debug mode (default: False).",
    )
    args = parser.parse_args()
    app.run(debug=args.debug, port=args.port)



# Copyright (c) 2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.Action import BaseAction
from grid2op.Agent.recoPowerlineAgent import RecoPowerlineAgent
from grid2op.Agent.baseAgent import BaseAgent
from grid2op.Observation import BaseObservation
from grid2op.dtypes import dt_int


class AlertAgent(BaseAgent):
    """
    This is a :class:`AlertAgent` example, which will attempt to reconnect powerlines and send alerts on the worst possible attacks: for each disconnected powerline
    that can be reconnected, it will simulate the effect of reconnecting it. And reconnect the one that lead to the
    highest simulated reward. It will also simulate the effect of having a line disconnection on attackable lines and raise alerts for the worst ones

    """

    def __init__(self,
                 action_space,
                 grid_controler=RecoPowerlineAgent,
                 percentage_alert=30,
                 simu_step=1,
                 threshold=0.99):
        super().__init__(action_space)
        if isinstance(grid_controler, type):
            self.grid_controler = grid_controler(action_space)
        else:
            self.grid_controler = grid_controler
            
        self.percentage_alert = percentage_alert
        self.simu_step = simu_step
        self.threshold = threshold  # if the max flow after a line disconnection is below threshold, then the alert is not raised
        
        # store the result of the simulation of powerline disconnection
        self.alertable_line_ids = type(action_space).alertable_line_ids
        self.n_alertable_lines = len(self.alertable_line_ids)
        self.nb_overloads = np.zeros(self.n_alertable_lines, dtype=dt_int)
        self.rho_max_N_1 = np.zeros(self.n_alertable_lines)
        self.N_1_actions = [self.action_space({"set_line_status": [(id_, -1)]}) for id_ in self.alertable_line_ids]
        self._first_k = np.zeros(self.n_alertable_lines, dtype=bool)
        self._first_k[:int(self.percentage_alert / 100. * self.n_alertable_lines)] = True
        
    def act(self, observation: BaseObservation, reward: float, done: bool = False) -> BaseAction:
        action = self.grid_controler.act(observation, reward, done)

        self.nb_overloads[:] = 0
        self.rho_max_N_1[:] = 0.
        
        # test which backend to know which method to call
        for i, tmp_act in enumerate(self.N_1_actions):
            # only simulate if the line is connected
            if observation.line_status[self.alertable_line_ids[i]]:
                action_to_simulate = tmp_act
                action_to_simulate += action
                action_to_simulate.remove_line_status_from_topo(observation)
                (
                    simul_obs,
                    simul_reward,
                    simul_done,
                    simul_info,
                ) = observation.simulate(action_to_simulate, time_step=self.simu_step)

                rho_simu = simul_obs.rho
                if not simul_done:
                    self.nb_overloads[i] = (rho_simu >= 1).sum()
                    self.rho_max_N_1[i] = (rho_simu).max()
                else:
                    self.nb_overloads[i] = type(observation).n_line
                    self.rho_max_N_1[i] = 5.
        
        # sort the index by nb_overloads and, if nb_overloads is equal, sort by rho_max
        ind = (self.nb_overloads * 1000. + self.rho_max_N_1).argsort()
        ind = ind[::-1]
        
        # send alerts when the powerline is among the top k (not to send too many alerts) and 
        # the max rho after the powerline disconnection is too high (above threshold)
        indices_to_keep = ind[self._first_k & (self.rho_max_N_1[ind] >= self.threshold)]
        action.raise_alert = [i for i in indices_to_keep]

        return action



# Copyright (c) 2019-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from typing import Tuple
import pandas as pd
import numpy as np

from grid2op.dtypes import dt_int, dt_float, dt_bool
from grid2op.Chronics.gridValue import GridValue
from grid2op.Chronics.handlers.csvHandler import CSVHandler


class CSVMaintenanceHandler(CSVHandler):
    """Reads and produce time series if given by a csv file (possibly compressed).
    
    The separator used can be specified as input. 
    
    The file name should match the "array_name". If you want to use
    the maintenance file present in the file "my_maintenance_file.csv.gz"
    then you should create a CSVMaintenanceHandler with 
    `array_name="my_maintenance_file"`.
    
    The csv should be structured as follow:
    
    - it should not have any "index" or anything, only data used by 
      grid2op will be used
    - Each element powerline is represented by a `column`.
    - It should have a header with the name of the powerlines that
      should match the one in the environment. For example 
      if "0_1_0" is the name of a powerline in your environment, 
      then a column should be called "0_1_0".
    - each time step is represented as a `row` and in order. For example
      (removing the header), row 1 (first row) will be step 1, row 2 will
      be step 2 etc.
    - only binary data (0 or 1) should be present in the file. No "bool", 
      no string etc.
      
    .. warning::
        Use this class only for the ENVIRONMENT data ("load_p", "load_q",
        "prod_p" or "prod_v") and not for maintenance (in this case
        use :class:`CSVMaintenanceHandler`) nor for 
        forecast (in this case use :class:`CSVForecastHandler`) 
    
    This is the default way to provide data to grid2op and its used for
    most l2rpn environments.
    
    """
    def __init__(self,
                 array_name="maintenance",
                 sep=";",
                 max_iter=-1) -> None:
        super().__init__(array_name, sep, None, max_iter)
        # None corresponds to "chunk_size" which is not supported for maintenance
    
    def _init_attrs(self, array):
        super()._init_attrs(array)
        n_line = self.array.shape[1]
        
        self.maintenance_time = (
            np.zeros(shape=(self.array.shape[0], n_line), dtype=dt_int)
            - 1
        )
        self.maintenance_duration = np.zeros(
            shape=(self.array.shape[0], n_line), dtype=dt_int
        )

        # test that with chunk size
        for line_id in range(n_line):
            self.maintenance_time[:, line_id] = GridValue.get_maintenance_time_1d(
                self.array[:, line_id]
            )
            self.maintenance_duration[
                :, line_id
            ] = GridValue.get_maintenance_duration_1d(self.array[:, line_id])

        # there are _maintenance and hazards only if the value in the file is not 0.
        self.array = self.array != 0.0
        self.array = self.array.astype(dt_bool)
    
    def load_next_maintenance(self) -> Tuple[np.ndarray, np.ndarray]:
        maint_time = 1 * self.maintenance_time[self.current_index, :]
        maint_duration = 1 * self.maintenance_duration[self.current_index, :]
        return maint_time, maint_duration
    
    def set_chunk_size(self, chunk_size):
        # skip the definition in CSVHandler to jump to the level "above"
        return super(CSVHandler, self).set_chunk_size(chunk_size)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import numpy as np

from packaging import version
try:
    from importlib.metadata import distribution
except ModuleNotFoundError:
    # not available in python 3.7
    from importlib_metadata import distribution
NUMPY_VERSION = version.parse(distribution('numpy').version)
_MAX_NUMPY_VERSION_NPINT = version.parse("1.20.0") 

dt_int = (
    np.int32
)  # dtype('int64') or dtype('int32') depending on platform => i force it to int32
dt_float = (
    np.float32
)  # dtype('float64') or dtype('float32') depending on platform  => i force it to float32

dt_bool = np.bool_  # mandatory for numpy >= 1.24

int_types = (int, dt_int, np.int64)
if NUMPY_VERSION <= _MAX_NUMPY_VERSION_NPINT:
    int_types = (int, dt_int, np.int64, np.int)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from collections import OrderedDict

import numpy as np
from grid2op.Converter.Converters import Converter
from grid2op.dtypes import dt_float, dt_int


class ToVect(Converter):
    """
    This converters allows to manipulate the vector representation of the actions and observations.

    In this converter:

    - `encoded_act` are numpy ndarray
    - `transformed_obs` are numpy ndarray
      (read more about these concepts by looking at the documentation of :class:`grid2op.Converter.Converters`)

    It is convertible to a gym representation (like the original action space) in the form of a spaces.Box
    representing a continuous action space (even though most component are probably discrete).
    Note that if converted to a gym space, it is unlikely the method "sample" will yield to valid results.
    Most of the time it should generate Ambiguous action that will not be handled by grid2op.

    **NB** the conversion to a gym space should be done thanks to the :class:`grid2op.Converter.GymActionSpace`.
    """

    def __init__(self, action_space):
        Converter.__init__(self, action_space)
        self.init_action_space = action_space
        self.__class__ = ToVect.init_grid(action_space)
        self.do_nothing_vect = action_space({}).to_vect()

        # for gym conversion
        self.__gym_action_space = None
        self.__dict_space = None
        self.__order_gym = None
        self.__dtypes_gym = None
        self.__shapes_gym = None

        self.__order_gym_2_me = None
        self.__order_me_2_gym = None

    def convert_obs(self, obs):
        """
        This converter will match the observation to a vector, using the
        :func:`grid2op.Observation.BaseObservation.to_vect`
        function.

        Parameters
        ----------
        obs: :class:`grid2op.Observation.Observation`
            The observation, that will be processed into a numpy ndarray vector.

        Returns
        -------
        transformed_obs: ``numpy.ndarray``
            The vector representation of the action.

        """
        return obs.to_vect()

    def convert_act(self, encoded_act):
        """
        In this converter `encoded_act` is a numpy ndarray. This function transforms it back to a valid action.

        Parameters
        ----------
        encoded_act: ``numpy.ndarray``
            The action, represented as a vector

        Returns
        -------
        regular_act: :class:`grid2op.Action.Action`
            The corresponding action transformed with the :func:`grid2op.Action.BaseAction.from_vect`.

        """
        res = self.__call__({})
        res.from_vect(encoded_act, check_legit=False)
        return res

    def _init_gym_converter(self, cls_gym):
        if self.__gym_action_space is None:
            # i do that not to duplicate the code of the low / high bounds
            gym_action_space = cls_gym(self.init_action_space)
            low = tuple()
            high = tuple()
            order_gym = []
            dtypes = []
            shapes = []
            sizes = []
            prev = 0
            for k, v in gym_action_space.spaces.items():
                order_gym.append(k)
                dtypes.append(v.dtype)
                if isinstance(v, cls_gym._MultiBinaryType):
                    low += tuple([0 for _ in range(v.n)])
                    high += tuple([1 for _ in range(v.n)])
                    my_size = v.n
                elif isinstance(v, cls_gym._BoxType):
                    low += tuple(v.low)
                    high += tuple(v.high)
                    my_size = v.low.shape[0]
                else:
                    raise RuntimeError(
                        "Impossible to convert this converter to gym. Type {} of data "
                        "encountered while only MultiBinary and Box are supported for now."
                    )
                shapes.append(my_size)
                sizes.append(np.arange(my_size) + prev)
                prev += my_size
            self.__gym_action_space = gym_action_space
            my_type = cls_gym._BoxType(low=np.array(low), high=np.array(high), dtype=dt_float)

            order_me = []
            _order_gym_2_me = np.zeros(my_type.shape[0], dtype=dt_int) - 1
            _order_me_2_gym = np.zeros(my_type.shape[0], dtype=dt_int) - 1
            for el in self.init_action_space.attr_list_vect:
                order_me.append(cls_gym.keys_grid2op_2_human[el])

            prev = 0
            order_gym = list(gym_action_space.spaces.keys())
            for id_me, nm_attr in enumerate(order_me):
                id_gym = order_gym.index(nm_attr)
                index_me = np.arange(shapes[id_gym]) + prev
                _order_gym_2_me[sizes[id_gym]] = index_me
                _order_me_2_gym[index_me] = sizes[id_gym]
                # self.__order_gym_2_me[this_gym_ind] = sizes[id_me]
                prev += shapes[id_gym]
            self.__order_gym_2_me = _order_gym_2_me
            self.__order_me_2_gym = _order_me_2_gym
            self.__dict_space = {"action": my_type}
            self.__order_gym = order_gym
            self.__dtypes_gym = dtypes
            self.__shapes_gym = shapes

    def get_gym_dict(self, cls_gym):
        """
        Convert this action space int a "gym" action space represented by a dictionary (spaces.Dict)
        This dictionary counts only one keys which is "action" and inside this action is the
        
        cls_gym represents either :class:`grid2op.gym_compat.LegacyGymActionSpace` or
        :class:`grid2op.gym_compat.GymnasiumActionSpace`
        """
        self._init_gym_converter(cls_gym)
        return self.__dict_space

    def convert_action_from_gym(self, gymlike_action):
        """
        Convert a gym-like action (ie a Ordered dictionary with one key being only "action") to an
        action compatible with this converter (in this case a vectorized action).
        """
        vect = gymlike_action["action"]
        return vect[self.__order_gym_2_me]

    def convert_action_to_gym(self, action):
        """
        Convert a an action of this converter (ie a numpy array) into an action that is usable with
        an open ai gym (ie a Ordered dictionary with one key being only "action")
        """
        res = OrderedDict({"action": action[self.__order_me_2_gym]})
        return res



# Copyright (c) 2019-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

__all__ = ["BaseHandler",
           "CSVHandler",
           "CSVForecastHandler",
           "DoNothingHandler",
           "CSVMaintenanceHandler",
           "JSONMaintenanceHandler",
           "PersistenceForecastHandler",
           "PerfectForecastHandler",
           "NoisyForecastHandler",
           "LoadQFromPHandler",
           ]

from .baseHandler import BaseHandler
from .csvHandler import CSVHandler
from .do_nothing_handler import DoNothingHandler
from .csvForecastHandler import CSVForecastHandler
from .csvMaintenanceHandler import CSVMaintenanceHandler
from .jsonMaintenanceHandler import JSONMaintenanceHandler
from .persitenceForecastHandler import PersistenceForecastHandler
from .perfectForecastHandler import PerfectForecastHandler
from .noisyForecastHandler import NoisyForecastHandler
from .load_q_from_p_handler import LoadQFromPHandler



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

"""
This module is the base module for all graphical representation of the :class:`grid2op.BaseObservation.BaseObservation`.

It allows, from the layout of the graph of the powergrid (*eg* the coordinates of each substation) the position of
each objects (powerline ends, loads and generators) and the position of the buses in case of "node splitting" (when
a substations is split into independent electrical buses).

"""
import cmath
import math
import numpy as np
import warnings
import copy

from grid2op.Space import GridObjects
from grid2op.Exceptions import PlotError


class BasePlot(GridObjects):
    """
    INTERNAL

    .. warning:: /!\\\\ This module is deprecated /!\\\\

        Prefer using the module `grid2op.PlotGrid

    Utility class that allows to compute the position of the objects of the powergrid.

    Deriving from this class allows to perform the display of the powergrid.

    **NB** this class only performs the computation of the position, but does not display anything.

    Attributes
    -----------

    observation_space: :class:`grid2op.Observation.HelperObservation`
        The observation space used.

    """

    def __init__(
        self,
        observation_space,
        substation_layout=None,
        radius_sub=20.0,
        load_prod_dist=70.0,
        bus_radius=6.0,
    ):

        warnings.warn(
            "This whole class has been deprecated. Use `grid2op.PlotGrid module instead`",
            category=DeprecationWarning,
        )
        if substation_layout is None:
            if observation_space.grid_layout is None:
                # if no layout is provided, and observation_space has no layout, then it fails
                raise PlotError(
                    "Impossible to use plotting abilities without specifying a layout (coordinates) "
                    "of the substations."
                )

            # if no layout is provided, use the one in the observation_space
            substation_layout = []
            for el in observation_space.name_sub:
                substation_layout.append(observation_space.grid_layout[el])

        if len(substation_layout) != observation_space.n_sub:
            raise PlotError(
                "You provided a layout with {} elements while there are {} substations on the powergrid. "
                "Your layout is invalid".format(
                    len(substation_layout), observation_space.n_sub
                )
            )
        GridObjects.__init__(self)
        self.init_grid(observation_space)

        self.observation_space = observation_space
        self._layout = {}
        self._layout["substations"] = self._get_sub_layout(substation_layout)

        self.radius_sub = radius_sub
        self.load_prod_dist = load_prod_dist  # distance between load and generator to the center of the substation
        self.bus_radius = bus_radius

        self.subs_elements = [None for _ in self.observation_space.sub_info]

        # get the element in each substation
        for sub_id in range(self.observation_space.sub_info.shape[0]):
            this_sub = {}
            objs = self.observation_space.get_obj_connect_to(substation_id=sub_id)

            for c_id in objs["loads_id"]:
                c_nm = self._get_load_name(sub_id, c_id)
                this_load = {}
                this_load["type"] = "load"
                this_load["sub_pos"] = self.observation_space.load_to_sub_pos[c_id]
                this_sub[c_nm] = this_load

            for g_id in objs["generators_id"]:
                g_nm = self._get_gen_name(sub_id, g_id)
                this_gen = {}
                this_gen["type"] = "gen"
                this_gen["sub_pos"] = self.observation_space.gen_to_sub_pos[g_id]
                this_sub[g_nm] = this_gen

            for lor_id in objs["lines_or_id"]:
                ext_id = self.observation_space.line_ex_to_subid[lor_id]
                l_nm = self._get_line_name(sub_id, ext_id, lor_id)
                this_line = {}
                this_line["type"] = "line"
                this_line["sub_pos"] = self.observation_space.line_or_to_sub_pos[lor_id]
                this_sub[l_nm] = this_line

            for lex_id in objs["lines_ex_id"]:
                or_id = self.observation_space.line_or_to_subid[lex_id]
                l_nm = self._get_line_name(or_id, sub_id, lex_id)
                this_line = {}
                this_line["type"] = "line"
                this_line["sub_pos"] = self.observation_space.line_ex_to_sub_pos[lex_id]
                this_sub[l_nm] = this_line
            self.subs_elements[sub_id] = this_sub
        self._compute_layout()

        self.col_line = None
        self.col_sub = None
        self.col_load = None
        self.col_gen = None
        self.default_color = None

    def plot_layout(self, fig=None, reward=None, done=None, timestamp=None):
        """
        .. warning:: /!\\\\ This module is deprecated /!\\\\

            Prefer using the module `grid2op.PlotGrid

        This function plot the layout of the grid, as well as the object. You will see the name of each elements and
        their id.
        """
        fig = self.init_fig(fig, reward, done, timestamp)
        # draw powerline
        lines = self._draw_powerlines(fig)
        # draw substation
        subs = self._draw_subs(fig)
        # draw loads
        loads = self._draw_loads(fig)
        # draw gens
        gens = self._draw_gens(fig)
        self._post_process_obs(
            fig,
            reward=None,
            done=None,
            timestamp=None,
            subs=subs,
            lines=lines,
            loads=loads,
            gens=gens,
            topos=[],
        )
        return fig

    def plot_info(
        self,
        fig=None,
        line_info=None,
        load_info=None,
        gen_info=None,
        sub_info=None,
        colormap=None,
        unit=None,
    ):

        """
        .. warning:: /!\\\\ This module is deprecated /!\\\\

            Prefer using the module `grid2op.PlotGrid

        Plot some information on the powergrid. For now, only numeric data are supported.

        Parameters
        ----------
        line_info: ``list``
            information to be displayed in the powerlines, in place of their name and id (for example their
            thermal limit) [must have the same size as the number of powerlines and convertible to float]

        load_info: ``list``
            information to display in the generators, in place of their name and id
            [must have the same size as the number of loads and convertible to float]

        gen_info: ``list``
            information to display in the generators, in place of their name and id (for example their pmax)
            [must have the same size as the number of generators and convertible to float]

        sub_info: ``list``
            information to display in the substation, in place of their name and id (for example the number of
            different topologies possible at this substation) [must have the same size as the number of substations,
            and convertible to float]

        colormap: ``str``
            If not None, one of "line", "load", "gen" or "sub". If None, default colors will be used for each
            elements (default color is the coloring of
            If not None, all elements will be black, and the selected element will be highlighted.

        fig: ``matplotlib figure``
            The figure on which to draw. It is created by the method if ``None``.

        unit: ``str``, optional
            The unit in which the data are provided. For example, if you provide in `line_info` some data in mega-watt
            (MW) you can add `unit="MW"` to have the unit display on the screen.

        """
        fig = self.init_fig(fig, reward=None, done=None, timestamp=None)

        # draw powerline
        unit_line = None
        if line_info is not None:
            unit_line = unit
            if len(line_info) != self.n_line:
                raise PlotError(
                    "Impossible to display these information on the powerlines: there are {} elements"
                    "provided while {} powerlines on this grid".format(
                        len(line_info), self.n_line
                    )
                )
            line_info = np.array(line_info).astype(np.float)
        line_info = [line_info, line_info, line_info]
        lines = self._draw_powerlines(
            fig, vals=line_info, colormap=colormap, unit=unit_line
        )

        # draw substation
        unit_sub = None
        if sub_info is not None:
            unit_sub = unit
            if len(sub_info) != self.n_sub:
                raise PlotError(
                    "Impossible to display these information on the substations: there are {} elements"
                    "provided while {} substations on this grid".format(
                        len(sub_info), self.n_sub
                    )
                )
            sub_info = np.array(sub_info).astype(np.float)
        subs = self._draw_subs(fig, vals=sub_info, colormap=colormap, unit=unit_sub)

        # draw loads
        unit_load = None
        if load_info is not None:
            unit_load = unit
            if len(load_info) != self.n_load:
                raise PlotError(
                    "Impossible to display these information on the loads: there are {} elements"
                    "provided while {} loads on this grid".format(
                        len(load_info), self.n_load
                    )
                )
            load_info = np.array(load_info).astype(np.float)
        loads = self._draw_loads(fig, vals=load_info, colormap=colormap, unit=unit_load)

        # draw gens
        unit_gen = None
        if gen_info is not None:
            unit_gen = unit
            if len(gen_info) != self.n_gen:
                raise PlotError(
                    "Impossible to display these information on the generators: there are {} elements"
                    "provided while {} generators on this grid".format(
                        len(gen_info), self.n_gen
                    )
                )
            gen_info = np.array(gen_info).astype(np.float)
        gens = self._draw_gens(fig, vals=gen_info, colormap=colormap, unit=unit_gen)

        self._post_process_obs(
            fig,
            reward=None,
            done=None,
            timestamp=None,
            subs=subs,
            lines=lines,
            loads=loads,
            gens=gens,
            topos=[],
        )
        return fig

    def plot_obs(
        self,
        observation,
        fig=None,
        reward=None,
        done=None,
        timestamp=None,
        line_info="rho",
        load_info="p",
        gen_info="p",
        colormap="line",
    ):
        """
        .. warning:: /!\\\\ This module is deprecated /!\\\\

            Prefer using the module `grid2op.PlotGrid

        Plot the given observation in the given figure.

        For now it represents information about load and generator active values.
        It also display dashed powerlines when they are disconnected and the color of each powerlines depends on
        its relative flow (its flow in amperes divided by its maximum capacity).

        If a substation counts only 1 bus, nothing specific is display. If it counts more, then buses are materialized
        by colored dot and lines will connect every object to its appropriate bus (with the proper color).

        Names of substation and objects are NOT displayed on this figure to lower the amount of information.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The observation to plot

        fig: :class:`plotly.graph_objects.Figure`
            The figure on which to plot the observation. Possibly ``None``, in this case a new figure is made.

        line_info: ``str``
            One of "rho", "a", or "p" or "v" the information that will be plotted on the powerline By default "rho".
            All flow are taken "origin" side.

        load_info: ``str``
            One of "p" or "v" the information displayed on the load (défault to "p").

        gen_info: ``str``
            One of "p" or "v" the information displayed on the generators (default to "p").

        Returns
        -------
        res: :class:`plotly.graph_objects.Figure`
            The figure updated with the data from the new observation.
        """
        fig = self.init_fig(fig, reward, done, timestamp)

        # draw substation
        subs = self._draw_subs(fig=fig, vals=[None for el in range(self.n_sub)])

        # draw powerlines
        if line_info == "rho":
            line_vals = [observation.rho]
            line_units = "%"
        elif line_info == "a":
            line_vals = [observation.a_or]
            line_units = "A"
        elif line_info == "p":
            line_vals = [observation.p_or]
            line_units = "MW"
        elif line_info == "v":
            line_vals = [observation.v_or]
            line_units = "kV"
        else:
            raise PlotError(
                'Impossible to plot value "{}" for line. Possible values are "rho", "p", "v" and "a".'
            )
        line_vals.append(observation.line_status)
        line_vals.append(observation.p_or)
        lines = self._draw_powerlines(
            fig, vals=line_vals, unit=line_units, colormap=colormap
        )

        # draw the loads
        if load_info == "p":
            loads_vals = -observation.load_p
            load_units = "MW"
        elif load_info == "v":
            loads_vals = observation.load_v
            load_units = "kV"
        else:
            raise PlotError(
                'Impossible to plot value "{}" for load. Possible values are "p" and "v".'
            )
        loads = self._draw_loads(
            fig, vals=loads_vals, unit=load_units, colormap=colormap
        )

        # draw the generators
        if gen_info == "p":
            gen_vals = observation.prod_p
            gen_units = "MW"
        elif gen_info == "v":
            gen_vals = observation.prod_v
            gen_units = "kV"
        else:
            raise PlotError(
                'Impossible to plot value "{}" for generators. Possible values are "p" and "v".'
            )
        gens = self._draw_gens(fig, vals=gen_vals, unit=gen_units, colormap=colormap)
        # draw the topologies
        topos = self._draw_topos(fig=fig, observation=observation)
        self._post_process_obs(
            fig, reward, done, timestamp, subs, lines, loads, gens, topos
        )
        return fig

    def _get_sub_layout(self, init_layout):
        return init_layout

    def _get_line_name(self, subor_id, sub_ex_id, line_id):
        l_nm = "l_{}_{}_{}".format(subor_id, sub_ex_id, line_id)
        return l_nm

    def _get_load_name(self, sub_id, c_id):
        c_nm = "load_{}_{}".format(sub_id, c_id)
        return c_nm

    def _get_gen_name(self, sub_id, g_id):
        p_nm = "gen_{}_{}".format(sub_id, g_id)
        return p_nm

    def _compute_layout(self):
        """

        .. warning:: /!\\\\ This module is deprecated /!\\\\

            Prefer using the module `grid2op.PlotGrid

        Compute the position of each of the objects.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The observation used to know which object belong where.

        Returns
        -------

        """
        self._layout["line"] = {}

        # assign powerline coordinates
        for line_id in range(self.n_line):
            if line_id not in self._layout["line"]:
                # state = observation.state_of(line_id=line_id)
                sub_or_id = self.line_or_to_subid[line_id]  # state["origin"]["sub_id"]
                sub_ex_id = self.line_ex_to_subid[
                    line_id
                ]  # state["extremity"]["sub_id"]
                pos_or = self._layout["substations"][sub_or_id]
                pos_ex = self._layout["substations"][sub_ex_id]

                # make sure the powerline are connected to the circle of the substation and not to the center of it
                z_or_tmp = pos_or[0] + 1j * pos_or[1]
                z_ex_tmp = pos_ex[0] + 1j * pos_ex[1]

                module_or = cmath.phase(z_ex_tmp - z_or_tmp)
                module_ex = cmath.phase(-(z_ex_tmp - z_or_tmp))

                # check parrallel lines:
                # for now it works only if there are 2 parrallel lines. The idea is to add / withdraw
                # 10° for each module in this case.
                # TODO draw line but not straight line in this case, this looks ugly for now :-/
                deg_parrallel = 25
                tmp_parrallel = self.observation_space.get_lines_id(
                    from_=sub_or_id, to_=sub_ex_id
                )
                if len(tmp_parrallel) > 1:
                    if line_id == tmp_parrallel[0]:
                        module_or += deg_parrallel / 360 * 2 * cmath.pi
                        module_ex -= deg_parrallel / 360 * 2 * cmath.pi
                    else:
                        module_or -= deg_parrallel / 360 * 2 * cmath.pi
                        module_ex += deg_parrallel / 360 * 2 * cmath.pi

                z_or = z_or_tmp + self.radius_sub * cmath.exp(module_or * 1j)
                z_ex = z_ex_tmp + self.radius_sub * cmath.exp(module_ex * 1j)
                pos_or = z_or.real, z_or.imag
                pos_ex = z_ex.real, z_ex.imag
                self._layout["line"][line_id] = sub_or_id, sub_ex_id
                # TODO here get proper name
                l_nm = self._get_line_name(sub_or_id, sub_ex_id, line_id)

                self.subs_elements[sub_or_id][l_nm]["pos"] = pos_or
                self.subs_elements[sub_or_id][l_nm]["z"] = z_or
                self.subs_elements[sub_ex_id][l_nm]["pos"] = pos_ex
                self.subs_elements[sub_ex_id][l_nm]["z"] = z_ex

        # assign loads and generators coordinates
        # this is done by first computing the "optimal" placement if there were only substation (so splitting equally
        # the objects around the circle) and then remove the closest position that are taken by the powerlines.
        for sub_id, elements in enumerate(self.subs_elements):
            nb_el = len(elements)

            # equally split
            pos_sub = self._layout["substations"][sub_id]
            z_sub = pos_sub[0] + 1j * pos_sub[1]
            pos_possible = [
                self.radius_sub * cmath.exp(1j * 2 * cmath.pi * i / nb_el) + z_sub
                for i in range(nb_el)
            ]

            # remove powerlines (already assigned)
            for el_nm, dict_el in elements.items():
                if dict_el["type"] == "line":
                    z = dict_el["z"]
                    closest = np.argmin([abs(pos - z) ** 2 for pos in pos_possible])
                    pos_possible = [
                        el for i, el in enumerate(pos_possible) if i != closest
                    ]

            i = 0
            # now assign load and generator
            for el_nm, dict_el in elements.items():
                if dict_el["type"] != "line":
                    dict_el["pos"] = (pos_possible[i].real, pos_possible[i].imag)
                    dict_el["z"] = pos_possible[i]
                    i += 1

        self._layout["load"] = {}
        for c_id in range(self.n_load):
            # state = observation.state_of(load_id=c_id)
            # sub_id = state["sub_id"]
            sub_id = self.load_to_subid[c_id]
            self._layout["load"][c_id] = sub_id

        self._layout["gen"] = {}
        for g_id in range(self.n_gen):
            # state = observation.state_of(gen_id=g_id)
            # sub_id = state["sub_id"]
            sub_id = self.gen_to_subid[g_id]
            self._layout["gen"][g_id] = sub_id

    def _get_line_coord(self, line_id):
        sub_or_id, sub_ex_id = self._layout["line"][line_id]
        l_nm = self._get_line_name(sub_or_id, sub_ex_id, line_id)
        pos_or = self.subs_elements[sub_or_id][l_nm]["pos"]
        pos_ex = self.subs_elements[sub_ex_id][l_nm]["pos"]
        return pos_or, pos_ex

    def _get_load_coord(self, load_id):
        sub_id = self._layout["load"][load_id]
        c_nm = self._get_load_name(sub_id, load_id)

        if not "elements_display" in self.subs_elements[sub_id][c_nm]:
            pos_load_sub = self.subs_elements[sub_id][c_nm]["pos"]
            pos_center_sub = self._layout["substations"][sub_id]

            z_sub = pos_center_sub[0] + 1j * pos_center_sub[1]
            theta = cmath.phase((self.subs_elements[sub_id][c_nm]["z"] - z_sub))
            pos_load = z_sub + cmath.exp(1j * theta) * self.load_prod_dist

            # position of the end of the line connecting the object to the substation
            pos_end_line = pos_load - cmath.exp(1j * theta) * 20
            how_center = self._get_position(theta)
            tmp_dict = {
                "pos_end_line": pos_end_line,
                "pos_load_sub": pos_load_sub,
                "pos_load": pos_load,
                "how_center": how_center,
            }
            self.subs_elements[sub_id][c_nm]["elements_display"] = tmp_dict
        else:
            dict_element = self.subs_elements[sub_id][c_nm]["elements_display"]
            pos_end_line = dict_element["pos_end_line"]
            pos_load_sub = dict_element["pos_load_sub"]
            pos_load = dict_element["pos_load"]
            how_center = dict_element["how_center"]

        return pos_end_line, pos_load_sub, pos_load, how_center

    def _get_gen_coord(self, gen_id):
        sub_id = self._layout["gen"][gen_id]
        c_nm = self._get_gen_name(sub_id, gen_id)

        if not "elements_display" in self.subs_elements[sub_id][c_nm]:
            pos_gen_sub = self.subs_elements[sub_id][c_nm]["pos"]
            pos_center_sub = self._layout["substations"][sub_id]

            z_sub = pos_center_sub[0] + 1j * pos_center_sub[1]
            theta = cmath.phase((self.subs_elements[sub_id][c_nm]["z"] - z_sub))
            pos_gen = z_sub + cmath.exp(1j * theta) * self.load_prod_dist

            # position of the end of the line connecting the object to the substation
            pos_end_line = pos_gen - cmath.exp(1j * theta) * 20
            how_center = self._get_position(theta)
            tmp_dict = {
                "pos_end_line": pos_end_line,
                "pos_gen_sub": pos_gen_sub,
                "pos_gen": pos_gen,
                "how_center": how_center,
            }
            self.subs_elements[sub_id][c_nm]["elements_display"] = tmp_dict
        else:
            dict_element = self.subs_elements[sub_id][c_nm]["elements_display"]
            pos_end_line = dict_element["pos_end_line"]
            pos_gen_sub = dict_element["pos_gen_sub"]
            pos_gen = dict_element["pos_gen"]
            how_center = dict_element["how_center"]

        return pos_end_line, pos_gen_sub, pos_gen, how_center

    def _get_topo_coord(self, sub_id, observation, elements):
        pos_center_sub = self._layout["substations"][sub_id]
        z_sub = pos_center_sub[0] + 1j * pos_center_sub[1]

        tmp = observation.state_of(substation_id=sub_id)
        if tmp["nb_bus"] == 1:
            # not to overload the plot, if everything is at the same bus, i don't plot it
            return [], []
        # I have at least 2 buses

        # I compute the position of each elements
        bus_vect = tmp["topo_vect"]

        # i am not supposed to have more than 2 buses
        buses_z = [None, None]  # center of the different buses
        nb_co = [0, 0]  # center of the different buses

        # the position of a bus is for now the average of all the elements in there
        for el_nm, dict_el in elements.items():
            this_el_bus = bus_vect[dict_el["sub_pos"]] - 1
            if this_el_bus >= 0:
                nb_co[this_el_bus] += 1
                if buses_z[this_el_bus] is None:
                    buses_z[this_el_bus] = dict_el["z"]
                else:
                    buses_z[this_el_bus] += dict_el["z"]

        #
        buses_z = [el / nb for el, nb in zip(buses_z, nb_co)]
        theta_z = [cmath.phase((el - z_sub)) for el in buses_z]

        # try to have nodes "in opposition" to one another
        NN = np.array(nb_co) / nb_co.sum()
        diff_theta = theta_z[0] - theta_z[1]
        # alpha = cmath.pi + diff_theta
        alpha = -cmath.pi + diff_theta
        alpha = math.fmod(alpha, 2 * cmath.pi)
        theta_z = [theta_z[0] - alpha * NN[1], theta_z[1] + alpha * NN[0]]

        # buses_z = [z_sub + (self.radius_sub - self.bus_radius) * 0.75 * cmath.exp(1j * theta) for theta in theta_z]
        buses_z = [
            z_sub + (self.radius_sub - self.bus_radius) * 0.6 * cmath.exp(1j * theta)
            for theta in theta_z
        ]
        return buses_z, bus_vect

    @staticmethod
    def _get_position(theta):
        quarter_pi = cmath.pi / 4
        half_pi = cmath.pi / 2.0
        if theta >= -quarter_pi and theta < quarter_pi:
            res = "center|left"
        elif theta >= quarter_pi and theta < quarter_pi + half_pi:
            res = "up|center"
        elif theta >= quarter_pi + half_pi and theta < quarter_pi + 2.0 * half_pi:
            res = "center|right"
        else:
            res = "down|center"
        return res

    def _get_text_unit(self, number, unit):
        if number is not None:
            if isinstance(number, float) or isinstance(number, np.float):
                if np.isfinite(number):
                    if unit == "%":
                        number *= 100.0
                    number = "{:.1f}".format(number)
                else:
                    return None

            if unit is not None:
                txt_ = "{}{}".format(number, unit)
            else:
                txt_ = number
        else:
            return None
        return txt_

    def _draw_subs(self, fig=None, vals=None, colormap=None, unit=None):
        subs = []
        colormap_ = lambda x: self.col_sub
        texts = None
        if vals is not None:
            texts = [self._get_text_unit(val, unit) for val in vals]

        if colormap is not None:
            if colormap == "sub":
                # normalize value for the color map
                vals = self._get_vals(vals)

        if texts is not None:
            vals = [float(text if text is not None else 0.0) for text in texts]

        for sub_id, center in enumerate(self._layout["substations"]):
            if texts is None:
                txt_ = "{}\nid: {}".format(self.name_sub[sub_id], sub_id)
                this_col = colormap_("")
            else:
                txt_ = texts[sub_id]
                if colormap == "sub":
                    this_col = self._get_sub_color_map(vals[sub_id])
                else:
                    this_col = self.default_color
            subs.append(self._draw_subs_one_sub(fig, sub_id, center, this_col, txt_))
        return subs

    def get_sub_color_map(self):
        return None

    def _draw_subs_one_sub(self, fig, sub_id, center, this_col, text):
        return None

    def _draw_powerlines(self, fig=None, vals=None, colormap=None, unit=None):
        lines = []

        colormap_ = lambda x: self.col_line
        texts = None
        if vals is not None:
            vals_0 = vals[0]
            texts = [self._get_text_unit(val, unit) for val in vals[0]]

        if colormap is not None:
            if colormap == "line" and unit != "%":
                # normalize the value for the color map
                vals_0 = self._get_vals(vals[0])

        for line_id in range(self.n_line):
            pos_or, pos_ex, *_ = self._get_line_coord(line_id)

            if texts is None:
                txt_ = "{}\nid: {}".format(self.name_line[line_id], line_id)
                this_col = colormap_("")
            else:
                txt_ = texts[line_id]
                if colormap == "line":
                    this_col = self._get_line_color_map(vals_0[line_id])
                else:
                    this_col = self.default_color

            if vals is not None:
                value = vals_0[line_id]
                status = vals[1][line_id]
                por = vals[2][line_id]
            else:
                value = 0.0
                status = True
                por = 1.0

            if por is None:
                por = 1.0
            if status is None:
                status = True

            if not status:
                this_col = self.default_color
            lines.append(
                self._draw_powerlines_one_powerline(
                    fig,
                    line_id,
                    pos_or,
                    pos_ex,
                    status,
                    value,
                    txt_,
                    por >= 0.0,
                    this_col,
                )
            )
        return lines

    def _draw_powerlines_one_powerline(
        self, fig, l_id, pos_or, pos_ex, status, value, txt_, or_to_ex, this_col
    ):
        return None

    def _draw_loads(self, fig=None, vals=None, colormap=None, unit=None):
        loads = []

        colormap_ = lambda x: self.col_load
        texts = None
        if vals is not None:
            texts = [self._get_text_unit(val, unit) for val in vals]

        if colormap is not None:
            if colormap == "load":
                # normalized the value for the color map
                vals = self._get_vals(vals)

        for c_id in range(self.n_load):
            pos_end_line, pos_load_sub, pos_load, how_center = self._get_load_coord(
                c_id
            )
            if texts is None:
                txt_ = "{}\nid: {}".format(self.name_load[c_id], c_id)
                this_col = colormap_("")
            else:
                txt_ = texts[c_id]
                if colormap == "load":
                    this_col = self._get_load_color_map(vals[c_id])
                else:
                    this_col = self.default_color

            loads.append(
                self._draw_loads_one_load(
                    fig,
                    c_id,
                    pos_load,
                    txt_,
                    pos_end_line,
                    pos_load_sub,
                    how_center,
                    this_col,
                )
            )
        return loads

    def _draw_loads_one_load(
        self,
        fig,
        l_id,
        pos_load,
        txt_,
        pos_end_line,
        pos_load_sub,
        how_center,
        this_col,
    ):
        return None

    def _get_sub_color_map(self, normalized_val):
        return self._get_default_cmap(normalized_val)

    def _get_load_color_map(self, normalized_val):
        return self._get_default_cmap(normalized_val)

    def _get_gen_color_map(self, normalized_val):
        return self._get_default_cmap(normalized_val)

    def _get_line_color_map(self, normalized_val):
        return self._get_default_cmap(normalized_val)

    def _get_default_cmap(self, normalized_val):
        return self.default_color

    def _get_vals(self, vals):
        vals = copy.deepcopy(vals)
        min_ = np.min(vals)
        max_ = np.max(vals)
        vals -= min_
        vals /= max_ - min_ + 1e-5
        # now vals is between 0 and 1, i push it toward 1 a bit to better see it
        vals += 0.5
        vals /= 1.5
        return vals

    def _draw_gens(self, fig=None, vals=None, colormap=None, unit=None):
        gens = []

        colormap_ = lambda x: self.col_gen
        texts = None
        if vals is not None:
            texts = [self._get_text_unit(val, unit) for val in vals]

        if colormap is not None:
            if colormap == "gen":
                # normalized the value for plot
                vals = self._get_vals(vals)

        for g_id in range(self.n_gen):
            pos_end_line, pos_gen_sub, pos_gen, how_center = self._get_gen_coord(g_id)
            if texts is None:
                txt_ = "{}\nid: {}".format(self.name_gen[g_id], g_id)
                this_col = colormap_("")
            else:
                txt_ = texts[g_id]
                if colormap == "gen":
                    this_col = self._get_gen_color_map(vals[g_id])
                else:
                    this_col = self.default_color
            gens.append(
                self._draw_gens_one_gen(
                    fig,
                    g_id,
                    pos_gen,
                    txt_,
                    pos_end_line,
                    pos_gen_sub,
                    how_center,
                    this_col,
                )
            )
        return gens

    def _draw_gens_one_gen(
        self, fig, g_id, pos_gen, txt_, pos_end_line, pos_gen_sub, how_center, this_col
    ):
        return None

    def _draw_topos(self, observation, fig):
        res_topo = []
        for sub_id, elements in enumerate(self.subs_elements):
            buses_z, bus_vect = self._get_topo_coord(sub_id, observation, elements)

            if not buses_z:
                # I don't plot details of substations with 1 bus for better quality
                continue
            res_topo += self._draw_topos_one_sub(
                fig, sub_id, buses_z, elements, bus_vect
            )
        return res_topo

    def _draw_topos_one_sub(self, fig, sub_id, buses_z, elements, bus_vect):
        return [None]

    def _post_process_obs(
        self, fig, reward, done, timestamp, subs, lines, loads, gens, topos
    ):
        pass

    def init_fig(self, fig, reward, done, timestamp):
        pass

    ## DEPRECATED FUNCTIONS
    def plot_observation(
        self, observation, fig=None, line_info="rho", load_info="p", gen_info="p"
    ):
        """

        .. warning:: /!\\\\ This module is deprecated /!\\\\

            Prefer using the module `grid2op.PlotGrid

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The observation to plot

        fig: ``plotly figure``
            will be created if ``None``

        line_info: ``str``
            One of "rho", "a", or "p" the information that will be plotted on the powerline By default "rho".

        load_info: ``str``
            One of "p" or "v" the information displayed on the load (défault to "p").

        gen_info: ``str``
            One of "p" or "v" the information displayed on the generators (default to "p").

        Returns
        -------
        res: ``plotly figure``
            The resulting figure.
        """
        warnings.warn(
            '"plot_observation" method will be deprecated in future version. '
            'Please use "plot_obs" instead.',
            category=PendingDeprecationWarning,
        )

        res = self.plot_obs(
            observation,
            fig=fig,
            line_info=line_info,
            load_info=load_info,
            gen_info=gen_info,
        )
        return res

    def get_plot_observation(
        self, observation, fig=None, line_info="rho", load_info="p", gen_info="p"
    ):
        """

        .. warning:: /!\\\\ This module is deprecated /!\\\\

            Prefer using the module `grid2op.PlotGrid

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The observation to plot

        fig: ``plotly figure``
            will be created if ``None``

        line_info: ``str``
            One of "rho", "a", or "p" the information that will be plotted on the powerline By default "rho".

        load_info: ``str``
            One of "p" or "v" the information displayed on the load (défault to "p").

        gen_info: ``str``
            One of "p" or "v" the information displayed on the generators (default to "p").

        Returns
        -------
        res: ``plotly figure``
            The resulting figure.
        """
        warnings.warn(
            '"get_plot_observation" method will be deprecated in future version. '
            'Please use "plot_obs" instead.',
            category=PendingDeprecationWarning,
        )

        res = self.plot_obs(
            observation,
            fig=fig,
            line_info=line_info,
            load_info=load_info,
            gen_info=gen_info,
        )
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import os
import warnings
import pkg_resources

from grid2op.Environment import Environment
from grid2op.Backend import Backend, PandaPowerBackend
from grid2op.Parameters import Parameters
from grid2op.Chronics import ChronicsHandler, Multifolder, ChangeNothing
from grid2op.Chronics import (
    GridStateFromFile,
    GridStateFromFileWithForecasts,
    GridValue,
)
from grid2op.Action import (
    BaseAction,
    TopologyAction,
    TopologyAndDispatchAction,
    DontAct,
)
from grid2op.Exceptions import *
from grid2op.Observation import CompleteObservation, BaseObservation
from grid2op.Reward import BaseReward, L2RPNReward, RedispReward
from grid2op.Rules import BaseRules, AlwaysLegal, DefaultRules
from grid2op.VoltageControler import ControlVoltageFromFile
from grid2op.Opponent import BaseOpponent

from grid2op.Chronics.Settings_L2RPN2019 import (
    L2RPN2019_CASEFILE,
    L2RPN2019_DICT_NAMES,
    ReadPypowNetData,
    CASE_14_L2RPN2019_LAYOUT,
)
from grid2op.Chronics.Settings_5busExample import (
    EXAMPLE_CHRONICSPATH,
    EXAMPLE_CASEFILE,
    CASE_5_GRAPH_LAYOUT,
)
from grid2op.Chronics.Settings_case14_test import (
    case14_test_CASEFILE,
    case14_test_CHRONICSPATH,
    case14_test_TH_LIM,
)
from grid2op.Chronics.Settings_case14_redisp import (
    case14_redisp_CASEFILE,
    case14_redisp_CHRONICSPATH,
    case14_redisp_TH_LIM,
)
from grid2op.Chronics.Settings_case14_realistic import (
    case14_real_CASEFILE,
    case14_real_CHRONICSPATH,
    case14_real_TH_LIM,
)
from grid2op.MakeEnv.get_default_aux import _get_default_aux


data_folder = pkg_resources.resource_filename("grid2op", "data")
CASE_14_FILE = os.path.abspath(
    os.path.join(data_folder, "rte_case14_redisp", "grid.json")
)
CHRONICS_FODLER = os.path.abspath(
    os.path.join(data_folder, "rte_case14_redisp", "chronics", "0")
)
CHRONICS_MLUTIEPISODE = os.path.join(data_folder, "rte_case14_redisp", "chronics")

NAMES_CHRONICS_TO_BACKEND = {
    "loads": {
        "2_C-10.61": "load_1_0",
        "3_C151.15": "load_2_1",
        "14_C63.6": "load_13_2",
        "4_C-9.47": "load_3_3",
        "5_C201.84": "load_4_4",
        "6_C-6.27": "load_5_5",
        "9_C130.49": "load_8_6",
        "10_C228.66": "load_9_7",
        "11_C-138.89": "load_10_8",
        "12_C-27.88": "load_11_9",
        "13_C-13.33": "load_12_10",
    },
    "lines": {
        "1_2_1": "0_1_0",
        "1_5_2": "0_4_1",
        "9_10_16": "8_9_2",
        "9_14_17": "8_13_3",
        "10_11_18": "9_10_4",
        "12_13_19": "11_12_5",
        "13_14_20": "12_13_6",
        "2_3_3": "1_2_7",
        "2_4_4": "1_3_8",
        "2_5_5": "1_4_9",
        "3_4_6": "2_3_10",
        "4_5_7": "3_4_11",
        "6_11_11": "5_10_12",
        "6_12_12": "5_11_13",
        "6_13_13": "5_12_14",
        "4_7_8": "3_6_15",
        "4_9_9": "3_8_16",
        "5_6_10": "4_5_17",
        "7_8_14": "6_7_18",
        "7_9_15": "6_8_19",
    },
    "prods": {
        "1_G137.1": "gen_0_4",
        "3_G36.31": "gen_2_1",
        "6_G63.29": "gen_5_2",
        "2_G-56.47": "gen_1_0",
        "8_G40.43": "gen_7_3",
    },
}

ALLOWED_KWARGS_MAKE = {
    "param",
    "backend",
    "observation_class",
    "gamerules_class",
    "chronics_path",
    "reward_class",
    "action_class",
    "grid_path",
    "names_chronics_to_backend",
    "data_feeding_kwargs",
    "chronics_class",
    "volagecontroler_class",
    "other_rewards",
    "opponent_action_class",
    "opponent_class",
    "opponent_init_budget",
}

ERR_MSG_KWARGS = {
    "backend": 'The backend of the environment (keyword "backend") must be an instance of grid2op.Backend',
    "observation_class": 'The type of observation of the environment (keyword "observation_class")'
    " must be a subclass of grid2op.BaseObservation",
    "param": 'The parameters of the environment (keyword "param") must be an instance of grid2op.Parameters',
    "gamerules_class": 'The type of rules of the environment (keyword "gamerules_class")'
    " must be a subclass of grid2op.BaseRules",
    "reward_class": 'The type of reward in the environment (keyword "reward_class") must be a subclass of grid2op.BaseReward',
    "action_class": 'The type of action of the environment (keyword "action_class") must be a subclass of grid2op.BaseAction',
    "data_feeding_kwargs": "The argument to build the data generation process [chronics]"
    '  (keyword "data_feeding_kwargs") should be a dictionnary.',
    "chronics_class": 'The argument to build the data generation process [chronics] (keyword "chronics_class")'
    " should be a class that inherit grid2op.Chronics.GridValue.",
    "chronics_handler": 'The argument to build the data generation process [chronics] (keyword "data_feeding")'
    " should be a class that inherit grid2op.ChronicsHandler.ChronicsHandler.",
    "voltagecontroler_class": 'The argument to build the online controler for chronics (keyword "volagecontroler_class")'
    " should be a class that inherit grid2op.VoltageControler.ControlVoltageFromFile.",
    "names_chronics_to_grid": 'The converter between names (keyword "names_chronics_to_backend") should be a dictionnary.',
    "other_rewards": 'The argument to build the online controler for chronics (keyword "other_rewards") '
    "should be dictionnary.",
    "opponent_action_class": 'The argument used to build the "opponent_action_class" should be a class that '
    'inherit from "BaseAction"',
    "opponent_class": 'The argument used to build the "opponent_class" should be a class that '
    'inherit from "BaseOpponent"',
    "opponent_init_budget": 'The initial budget of the opponent "opponent_init_budget" should be a float',
    "chronics_path": 'The path where the data is located (keyword "chronics_path") should be a string.',
    "grid_path": 'The path where the grid is located (keyword "grid_path") should be a string.',
}


def make_old(name_env="case14_realistic", **kwargs):
    """
    INTERNAL USE ONLY

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    (DEPRECATED) This function is a shortcut to rapidly create some (pre defined) environments within the grid2op Framework.

    For now, only the environment corresponding to the IEEE "case14" powergrid, with some pre defined chronics
    is available.

    Other environments, with different powergrids will be made available in the future.

    It mimic the ``gym.make`` function.

    Parameters
    ----------
    name_env: ``str``
        Name of the environment to create.

    param: ``grid2op.Parameters.Parameters``, optional
        Type of parameters used for the Environment. Parameters defines how the powergrid problem is cast into an
        markov decision process, and some internal

    backend: ``grid2op.Backend.Backend``, optional
        The backend to use for the computation. If provided, it must be an instance of :class:`grid2op.Backend.Backend`.

    action_class: ``type``, optional
        Type of BaseAction the BaseAgent will be able to perform.
        If provided, it must be a subclass of :class:`grid2op.BaseAction.BaseAction`

    observation_class: ``type``, optional
        Type of BaseObservation the BaseAgent will receive.
        If provided, It must be a subclass of :class:`grid2op.BaseAction.BaseObservation`

    reward_class: ``type``, optional
        Type of reward signal the BaseAgent will receive.
        If provided, It must be a subclass of :class:`grid2op.BaseReward.BaseReward`

    gamerules_class: ``type``, optional
        Type of "Rules" the BaseAgent need to comply with. Rules are here to model some operational constraints.
        If provided, It must be a subclass of :class:`grid2op.RulesChecker.BaseRules`

    grid_path: ``str``, optional
        The path where the powergrid is located.
        If provided it must be a string, and point to a valid file present on the hard drive.

    data_feeding_kwargs: ``dict``, optional
        Dictionnary that is used to build the `data_feeding` (chronics) objects.

    chronics_class: ``type``, optional
        The type of chronics that represents the dynamics of the Environment created. Usually they come from different
        folders.

    data_feeding: ``type``, optional
        The type of chronics handler you want to use.

    chronics_path: ``str``
        Path where to look for the chronics dataset.

    volagecontroler_class: ``type``, optional
        The type of :class:`grid2op.VoltageControler.VoltageControler` to use, it defaults to

    other_rewards: ``dict``, optional
        Dictionnary with other rewards we might want to look at at during training. It is given as a dictionnary with
        keys the name of the reward, and the values a class representing the new variables.

    Returns
    -------
    env: :class:`grid2op.Environment.Environment`
        The created environment.
    """

    warnings.warn("make_old is deprecated. Please consider using make instead")
    for el in kwargs:
        if not el in ALLOWED_KWARGS_MAKE:
            raise EnvError(
                'Unknown keyword argument "{}" used to create an Environement. '
                "No Environment will be created. "
                "Accepted keyword arguments are {}".format(
                    el, sorted(ALLOWED_KWARGS_MAKE)
                )
            )

    # first extract parameters that doesn't not depend on the powergrid

    ## the parameters of the game, thermal limits threshold, simulate cascading failure, powerflow mode etc. (the gamification of the game)
    msg_error = 'The parameters of the environment (keyword "param") must be an instance of grid2op.Parameters'
    param = _get_default_aux(
        "param",
        kwargs,
        defaultClass=Parameters,
        defaultClassApp=Parameters,
        msg_error=msg_error,
    )

    ## the backend use, to compute the powerflow
    msg_error = 'The backend of the environment (keyword "backend") must be an instance of grid2op.Backend'
    backend = _get_default_aux(
        "backend",
        kwargs,
        defaultClass=PandaPowerBackend,
        defaultClassApp=Backend,
        msg_error=msg_error,
    )

    ## type of observation the agent will receive
    msg_error = (
        'The type of observation of the environment (keyword "observation_class")'
    )
    msg_error += " must be a subclass of grid2op.BaseObservation"
    observation_class = _get_default_aux(
        "observation_class",
        kwargs,
        defaultClass=CompleteObservation,
        defaultClassApp=BaseObservation,
        msg_error=msg_error,
        isclass=True,
    )

    ## type of rules of the game (mimic the operationnal constraints)
    msg_error = 'The path where the data is located (keyword "chronics_path") should be a string.'
    chronics_path = _get_default_aux(
        "chronics_path",
        kwargs,
        defaultClassApp=str,
        defaultinstance="",
        msg_error=msg_error,
    )

    # bulid the default parameters for each case file
    data_feeding_default_class = ChronicsHandler
    gamerules_class = AlwaysLegal
    defaultinstance_chronics_kwargs = {}
    if name_env.lower() == "case14_fromfile":
        default_grid_path = CASE_14_FILE
        if chronics_path == "":
            chronics_path = CHRONICS_MLUTIEPISODE

        defaultinstance_chronics_kwargs = {
            "chronicsClass": Multifolder,
            "path": chronics_path,
            "gridvalueClass": GridStateFromFileWithForecasts,
        }
        default_name_converter = {}
        default_action_class = TopologyAction
        default_reward_class = L2RPNReward
    elif name_env.lower() == "l2rpn_2019":
        warnings.warn(
            'You are using the "l2rpn_2019" environmnet, which will be remove from this package in '
            'future versions. Please use "make_new" to download the real l2rpn dataset.'
        )
        if chronics_path == "":
            msg_error = "Default chronics (provided in this package) cannot be used with the environment "
            msg_error += (
                '"l2rpn_2019". Please download the training data using either the method described in'
                "Grid2Op/l2rpn_2019/README.md (if you downloaded the github repository) or\n"
                "running the command line script (in a terminal):\n"
                'python -m grid2op.download --name "l2rpn_2019" --path_save PATH\WHERE\YOU\WANT\TO\DOWNLOAD'
            )
            raise EnvError(msg_error)
        default_grid_path = L2RPN2019_CASEFILE
        defaultinstance_chronics_kwargs = {
            "chronicsClass": Multifolder,
            "path": chronics_path,
            "gridvalueClass": ReadPypowNetData,
        }
        default_name_converter = L2RPN2019_DICT_NAMES
        default_action_class = TopologyAction
        default_reward_class = L2RPNReward
        gamerules_class = DefaultRules
    elif name_env.lower() == "case5_example":
        if chronics_path == "":
            chronics_path = EXAMPLE_CHRONICSPATH

        default_grid_path = EXAMPLE_CASEFILE
        defaultinstance_chronics_kwargs = {
            "chronicsClass": Multifolder,
            "path": chronics_path,
            "gridvalueClass": GridStateFromFileWithForecasts,
        }
        default_name_converter = {}
        default_action_class = TopologyAction
        default_reward_class = L2RPNReward
        gamerules_class = DefaultRules
    elif name_env.lower() == "case14_test":
        if chronics_path == "":
            chronics_path = case14_test_CHRONICSPATH
            warnings.warn(
                'Your are using a case designed for testing purpose. Consider using the "case14_redisp" '
                "environment instead."
            )

        default_grid_path = case14_test_CASEFILE
        defaultinstance_chronics_kwargs = {
            "chronicsClass": Multifolder,
            "path": chronics_path,
            "gridvalueClass": GridStateFromFileWithForecasts,
        }
        default_name_converter = {}
        default_action_class = TopologyAndDispatchAction
        default_reward_class = RedispReward
        gamerules_class = DefaultRules
    elif name_env.lower() == "case14_redisp":
        if chronics_path == "":
            chronics_path = case14_redisp_CHRONICSPATH
            warnings.warn(
                "Your are using only 2 chronics for this environment. More can be download by running, "
                "from a command line:\n"
                'python -m grid2op.download --name "case14_redisp" '
                "--path_save PATH\WHERE\YOU\WANT\TO\DOWNLOAD\DATA"
            )

        default_grid_path = case14_redisp_CASEFILE
        defaultinstance_chronics_kwargs = {
            "chronicsClass": Multifolder,
            "path": chronics_path,
            "gridvalueClass": GridStateFromFileWithForecasts,
        }
        default_name_converter = {}
        default_action_class = TopologyAndDispatchAction
        default_reward_class = RedispReward
        gamerules_class = DefaultRules
    elif name_env.lower() == "case14_realistic":
        if chronics_path == "":
            chronics_path = case14_real_CHRONICSPATH
            warnings.warn(
                "Your are using only 2 chronics for this environment. More can be download by running, "
                "from a command line:\n"
                'python -m grid2op.download --name "case14_realistic" '
                "--path_save PATH\WHERE\YOU\WANT\TO\DOWNLOAD\DATA"
            )

        default_grid_path = case14_real_CASEFILE
        defaultinstance_chronics_kwargs = {
            "chronicsClass": Multifolder,
            "path": chronics_path,
            "gridvalueClass": GridStateFromFileWithForecasts,
        }
        default_name_converter = {}
        default_action_class = TopologyAndDispatchAction
        default_reward_class = RedispReward
        gamerules_class = DefaultRules
    elif name_env.lower() == "blank":
        default_name_converter = {}
        default_grid_path = ""
        default_action_class = TopologyAction
        default_reward_class = L2RPNReward
        gamerules_class = AlwaysLegal
    else:
        raise UnknownEnv(
            'Unknown Environment named "{}". Current known environments are "case14_fromfile" '
            '(default), "case5_example", "case14_redisp", "case14_realistic" '
            'and "l2rpn_2019"'.format(name_env)
        )

    if "chronicsClass" not in defaultinstance_chronics_kwargs:
        defaultinstance_chronics_kwargs["chronicsClass"] = ChangeNothing

    # extract powergrid dependant parameters
    ## type of rules of the game (mimic the operationnal constraints)
    msg_error = 'The type of rules of the environment (keyword "gamerules_class")'
    msg_error += " must be a subclass of grid2op.BaseRules"
    gamerules_class = _get_default_aux(
        "gamerules_class",
        kwargs,
        defaultClass=gamerules_class,
        defaultClassApp=BaseRules,
        msg_error=msg_error,
        isclass=True,
    )

    ## type of reward the agent will receive
    msg_error = 'The type of observation of the environment (keyword "reward_class")'
    msg_error += " must be a subclass of grid2op.BaseReward"
    reward_class = _get_default_aux(
        "reward_class",
        kwargs,
        defaultClass=default_reward_class,
        defaultClassApp=BaseReward,
        msg_error=msg_error,
        isclass=True,
    )

    ## type of action the BaseAgent can perform
    msg_error = 'The type of action of the environment (keyword "action_class") must be a subclass of grid2op.BaseAction'
    action_class = _get_default_aux(
        "action_class",
        kwargs,
        defaultClass=default_action_class,
        defaultClassApp=BaseAction,
        msg_error=msg_error,
        isclass=True,
    )

    ## the powergrid path to use
    msg_error = (
        'The path where the grid is located (keyword "grid_path") should be a string.'
    )
    grid_path = _get_default_aux(
        "grid_path",
        kwargs,
        defaultClassApp=str,
        defaultinstance=default_grid_path,
        msg_error=msg_error,
    )

    ##
    msg_error = 'The converter between names (keyword "names_chronics_to_backend") should be a dictionnary.'
    names_chronics_to_backend = _get_default_aux(
        "names_chronics_to_backend",
        kwargs,
        defaultClassApp=dict,
        defaultinstance=default_name_converter,
        msg_error=msg_error,
    )

    ## the chronics to use
    ### the arguments used to build the data, note that the arguments must be compatible with the chronics class
    msg_error = 'The argument to build the data generation process [chronics] (keyword "data_feeding_kwargs")'
    msg_error += " should be a dictionnary."
    data_feeding_kwargs = _get_default_aux(
        "data_feeding_kwargs",
        kwargs,
        defaultClassApp=dict,
        defaultinstance=defaultinstance_chronics_kwargs,
        msg_error=msg_error,
    )
    for el in defaultinstance_chronics_kwargs:
        if not el in data_feeding_kwargs:
            data_feeding_kwargs[el] = defaultinstance_chronics_kwargs[el]

    ### the chronics generator
    msg_error = 'The argument to build the data generation process [chronics] (keyword "chronics_class")'
    msg_error += " should be a class that inherit grid2op.ChronicsHandler.GridValue."
    chronics_class_used = _get_default_aux(
        "chronics_class",
        kwargs,
        defaultClassApp=GridValue,
        defaultClass=data_feeding_kwargs["chronicsClass"],
        msg_error=msg_error,
        isclass=True,
    )
    data_feeding_kwargs["chronicsClass"] = chronics_class_used

    ### the chronics generator
    msg_error = 'The argument to build the data generation process [chronics] (keyword "data_feeding")'
    msg_error += (
        " should be a class that inherit grid2op.ChronicsHandler.ChronicsHandler."
    )
    data_feeding = _get_default_aux(
        "data_feeding",
        kwargs,
        defaultClassApp=ChronicsHandler,
        defaultClass=data_feeding_default_class,
        build_kwargs=data_feeding_kwargs,
        msg_error=msg_error,
    )

    ### controler for voltages
    msg_error = 'The argument to build the online controler for chronics (keyword "volagecontroler_class")'
    msg_error += " should be a class that inherit grid2op.VoltageControler.ControlVoltageFromFile."
    volagecontroler_class = _get_default_aux(
        "volagecontroler_class",
        kwargs,
        defaultClassApp=ControlVoltageFromFile,
        defaultClass=ControlVoltageFromFile,
        msg_error=msg_error,
        isclass=True,
    )

    ### other rewards
    msg_error = 'The argument to build the online controler for chronics (keyword "other_rewards")'
    msg_error += " should be dictionnary."
    other_rewards = _get_default_aux(
        "other_rewards",
        kwargs,
        defaultClassApp=dict,
        defaultinstance={},
        msg_error=msg_error,
        isclass=False,
    )

    # Opponent
    opponent_action_class = _get_default_aux(
        "opponent_action_class",
        kwargs,
        defaultClassApp=BaseAction,
        defaultClass=DontAct,
        msg_error=ERR_MSG_KWARGS["opponent_action_class"],
        isclass=True,
    )
    opponent_class = _get_default_aux(
        "opponent_class",
        kwargs,
        defaultClassApp=BaseOpponent,
        defaultClass=BaseOpponent,
        msg_error=ERR_MSG_KWARGS["opponent_class"],
        isclass=True,
    )
    opponent_init_budget = _get_default_aux(
        "opponent_init_budget",
        kwargs,
        defaultClassApp=float,
        defaultinstance=0.0,
        msg_error=ERR_MSG_KWARGS["opponent_init_budget"],
        isclass=False,
    )
    if not os.path.exists(grid_path):
        raise EnvError(
            'There is noting at "{}" where the powergrid should be located'.format(
                os.path.abspath(grid_path)
            )
        )

    env = Environment(
        init_grid_path=grid_path,
        chronics_handler=data_feeding,
        backend=backend,
        parameters=param,
        names_chronics_to_backend=names_chronics_to_backend,
        actionClass=action_class,
        observationClass=observation_class,
        rewardClass=reward_class,
        legalActClass=gamerules_class,
        voltagecontrolerClass=volagecontroler_class,
        other_rewards=other_rewards,
        opponent_action_class=opponent_action_class,
        opponent_class=opponent_class,
        opponent_init_budget=opponent_init_budget,
        name=name_env,
    )

    # update the thermal limit if any
    if name_env.lower() == "case14_test":
        env.set_thermal_limit(case14_test_TH_LIM)
        env.attach_layout(CASE_14_L2RPN2019_LAYOUT)
    elif name_env.lower() == "case14_redisp":
        env.set_thermal_limit(case14_redisp_TH_LIM)
        env.attach_layout(CASE_14_L2RPN2019_LAYOUT)
    elif name_env.lower() == "case14_realistic":
        env.set_thermal_limit(case14_real_TH_LIM)
        env.attach_layout(CASE_14_L2RPN2019_LAYOUT)
    elif name_env.lower() == "l2rpn_2019":
        env.attach_layout(CASE_14_L2RPN2019_LAYOUT)
    elif name_env.lower() == "case5_example":
        env.attach_layout(CASE_5_GRAPH_LAYOUT)
    return env



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import io
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.path import Path

from grid2op.PlotGrid.BasePlot import BasePlot
from grid2op.PlotGrid.PlotUtil import PlotUtil as pltu
import matplotlib.patches as patches
from matplotlib.lines import Line2D

from grid2op.PlotGrid.config import *  # all colors


class GenDraw(patches.CirclePolygon):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    Empty class to handle the legend
    """

    def __init__(self, *args, resolution=5, **kwargs):
        patches.CirclePolygon.__init__(self, *args, resolution=resolution, **kwargs)


class LoadDraw(patches.CirclePolygon):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    Empty class to handle the legend
    """

    def __init__(self, *args, resolution=3, **kwargs):
        patches.CirclePolygon.__init__(self, *args, resolution=resolution, **kwargs)


class StorageDraw(patches.CirclePolygon):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    Empty class to handle the legend
    """

    def __init__(self, *args, resolution=4, **kwargs):
        patches.CirclePolygon.__init__(self, *args, resolution=resolution, **kwargs)


# TODO refactor this class to make possible some calls like
# plotmatplot.plot_info(...).plot_gentype(...) is possible

# TODO add some transparency when coloring=... is used in plot_info
# TODO code the load part in the plot_info


class PlotMatplot(BasePlot):
    """
    This class uses the python library "matplotlib" to draw the powergrid.

    Attributes
    ----------

    width: ``int``
        Width of the figure in pixels
    height: ``int``
        Height of the figure in pixel
    dpi: ``int``
        Dots per inch, to convert pixels dimensions into inches
    _scale: ``float``
        Scale of the drawing in arbitrary units
    _sub_radius: ``int``
        Substation circle size
    _sub_face_color: ``str``
        Substation circle fill color
    _sub_edge_color: ``str``
        Substation circle edge color
    _sub_txt_color: ``str``
        Substation info text color
    _load_radius: ``int``
        Load circle size
    _load_name: ``bool``
        Show load names (default True)
    _load_face_color: ``str``
        Load circle fill color
    _load_edge_color: ``str``
        Load circle edge color
    _load_txt_color: ``str``
        Load info text color
    _load_line_color: ``str``
        Color of the line from load to substation
    _load_line_width: ``int``
        Width of the line from load to substation
    _gen_radius: ``int``
        Generators circle size
    _gen_name: ``bool``
        Show generators names (default True)
    _gen_face_color: ``str``
        Generators circle fill color
    _gen_edge_color: ``str``
        Generators circle edge color
    _gen_txt_color: ``str``
        Generators info txt color
    _gen_line_color: ``str``
        Color of the line form generator to substation
    _gen_line_width: ``str``
        Width of the line from generator to substation
    _line_color_scheme: ``list``
        List of color strings to color powerlines based on rho values
    _line_color_width: ``int``
        Width of the powerlines lines
    _line_bus_radius: ``int``
        Size of the bus display circle
    _line_bus_face_colors: ``list``
        List of 3 colors strings, each corresponding to the fill color of the bus circle
    _line_arrow_len: ``int``
        Length of the arrow on the powerlines
    _line_arrow_width: ``int``
       Width of the arrow on the powerlines

    Examples
    --------
    You can use it this way:

    .. code-block:: python

        import grid2op
        from grid2op.PlotGrid import PlotMatplot
        env = grid2op.make()
        plot_helper = PlotMatplot(env.observation_space)

        # and now plot an observation (for example)
        obs = env.reset()
        fig = plot_helper.plot_obs(obs)
        fig.show()

        # more information about it on the `getting_started/8_PlottingCapabilities.ipynb` notebook of grid2op

    """

    def __init__(
        self,
        observation_space,
        width=1280,
        height=720,
        grid_layout=None,
        dpi=96,
        scale=2000.0,
        bus_radius=6,
        sub_radius=15,
        load_radius=8,
        load_name=False,
        load_id=False,
        load_resolution=3,  # number of edges of the polygon representing the generator
        gen_radius=8,
        gen_name=False,
        gen_id=False,
        gen_resolution=5,  # number of edges of the polygon representing the generator
        storage_resolution=4,  # number of edges of the polygon representing the generator
        line_name=False,
        line_id=False,
    ):
        self.dpi = dpi
        super().__init__(observation_space, width, height, scale, grid_layout)

        self._sub_radius = sub_radius
        self._sub_face_color = "w"
        self._sub_edge_color = "blue"
        self._sub_txt_color = "black"
        self._display_sub_name = True

        self._load_radius = load_radius
        self._load_name = load_name
        self._load_id = load_id
        self._load_face_color = "w"
        self._load_edge_color = "orange"
        self._load_resolution = load_resolution
        self._load_patch = self._load_patch_default
        self._load_txt_color = "black"
        self._load_line_color = "black"
        self._load_line_width = 1
        self._display_load_name = True

        self._gen_radius_orig = gen_radius
        self._gen_radius = None  # init in self.restore_gen_palette()
        self._gen_resolution = gen_resolution
        self._gen_patch = self._gen_patch_default
        self._gen_name = gen_name
        self._gen_id = gen_id
        self._gen_face_color = "w"
        self._gen_edge_color_orig = "green"
        self._gen_edge_color = None
        self._gen_txt_color = "black"
        self._gen_line_color = "black"
        self._gen_line_width_orig = 1
        self._gen_line_width = None
        self._display_gen_value = True
        self._display_gen_name = True
        self.restore_gen_palette()

        self._storage_radius = load_radius
        self._storage_name = load_name
        self._storage_id = load_id  # bool : do i plot the id
        self._storage_face_color = "w"
        self._storage_edge_color = "purple"
        self._storage_resolution = storage_resolution
        self._storage_patch = self._storage_patch_default
        self._storage_txt_color = "black"
        self._storage_line_color = "black"
        self._storage_line_width = 1
        self._display_storage_name = True

        self._line_name = line_name
        self._line_id = line_id
        self._line_color_scheme_orig = ["blue", "orange", "red"]
        self._line_color_scheme = None
        self.restore_line_palette()

        self._line_color_width = 1
        self._line_bus_radius = bus_radius
        self._line_bus_face_colors = ["black", "red", "lime"]
        self._line_arrow_len = 10
        self._line_arrow_width = 10.0

        self.xlim = [0, 0]
        self.xpad = 5
        self.ylim = [0, 0]
        self.ypad = 5

        # for easize manipulation
        self.legend = None
        self.figure = None

    def _gen_patch_default(self, xy, radius, edgecolor, facecolor):
        """default patch used to draw generator"""
        # TODO maybe make a better version of this
        patch = GenDraw(
            xy,
            radius=radius,
            edgecolor=edgecolor,
            facecolor=facecolor,
            resolution=self._gen_resolution,
            linewidth=self._gen_line_width,
        )
        return patch

    def _load_patch_default(self, xy, radius, edgecolor, facecolor):
        """default patch used to draw generator"""
        # TODO maybe make a better version of this
        patch = LoadDraw(
            xy,
            radius=radius,
            edgecolor=edgecolor,
            facecolor=facecolor,
            resolution=self._load_resolution,
        )
        return patch

    def _storage_patch_default(self, xy, radius, edgecolor, facecolor):
        """default patch used to draw generator"""
        # TODO maybe make a better version of this
        patch = StorageDraw(
            xy,
            radius=radius,
            edgecolor=edgecolor,
            facecolor=facecolor,
            resolution=self._storage_resolution,
        )
        return patch

    def _v_textpos_from_dir(self, dirx, diry):
        if diry > 0:
            return "bottom"
        else:
            return "top"

    def _h_textpos_from_dir(self, dirx, diry):
        if dirx == 0:
            return "center"
        elif dirx > 0:
            return "left"
        else:
            return "right"

    def create_figure(self):
        # lazy loading of graphics library (reduce loading time)
        # and mainly because matplolib has weird impact on argparse
        import matplotlib.pyplot as plt

        w_inch = self.width / self.dpi
        h_inch = self.height / self.dpi
        f = plt.figure(figsize=(w_inch, h_inch), dpi=self.dpi)
        self.ax = f.subplots()
        f.canvas.draw()
        return f

    def clear_figure(self, figure):
        self.xlim = [0, 0]
        self.ylim = [0, 0]
        figure.clear()
        self.ax = figure.subplots()

    def convert_figure_to_numpy_HWC(self, figure):
        w, h = figure.get_size_inches() * figure.dpi
        w = int(w)
        h = int(h)
        buf = io.BytesIO()
        figure.canvas.print_raw(buf)
        buf.seek(0)
        img_arr = np.frombuffer(buf.getvalue(), dtype=np.uint8)
        buf.close()
        img_arr = np.reshape(img_arr, (h, w, 4))
        return img_arr

    def _draw_substation_txt(self, pos_x, pos_y, text):
        self.ax.text(
            pos_x,
            pos_y,
            text,
            color=self._sub_txt_color,
            horizontalalignment="center",
            verticalalignment="center",
        )

    def _draw_substation_circle(self, pos_x, pos_y):
        patch = patches.Circle(
            (pos_x, pos_y),
            radius=self._sub_radius,
            facecolor=self._sub_face_color,
            edgecolor=self._sub_edge_color,
        )
        self.ax.add_patch(patch)

    def draw_substation(self, figure, observation, sub_id, sub_name, pos_x, pos_y):
        self.xlim[0] = min(self.xlim[0], pos_x - self._sub_radius)
        self.xlim[1] = max(self.xlim[1], pos_x + self._sub_radius)
        self.ylim[0] = min(self.ylim[0], pos_y - self._sub_radius)
        self.ylim[1] = max(self.ylim[1], pos_y + self._sub_radius)

        self._draw_substation_circle(pos_x, pos_y)
        if self._display_sub_name:
            self._draw_substation_txt(pos_x, pos_y, str(sub_id))

    def _draw_load_txt(self, pos_x, pos_y, sub_x, sub_y, text):
        dir_x, dir_y = pltu.vec_from_points(sub_x, sub_y, pos_x, pos_y)
        off_x, off_y = pltu.norm_from_points(sub_x, sub_y, pos_x, pos_y)
        txt_x = pos_x + off_x * self._gen_radius
        txt_y = pos_y + off_y * self._gen_radius
        ha = self._h_textpos_from_dir(dir_x, dir_y)
        va = self._v_textpos_from_dir(dir_x, dir_y)
        self.ax.text(
            txt_x,
            txt_y,
            text,
            color=self._load_txt_color,
            horizontalalignment=ha,
            fontsize="small",
            verticalalignment=va,
        )

    def _draw_load_name(self, pos_x, pos_y, txt):
        self.ax.text(
            pos_x,
            pos_y,
            txt,
            color=self._load_txt_color,
            va="center",
            ha="center",
            fontsize="x-small",
        )

    def _draw_load_circle(self, pos_x, pos_y):
        patch = self._load_patch(
            (pos_x, pos_y),
            radius=self._load_radius,
            facecolor=self._load_face_color,
            edgecolor=self._load_edge_color,
        )
        self.ax.add_patch(patch)

    def _draw_load_line(self, pos_x, pos_y, sub_x, sub_y):
        codes = [Path.MOVETO, Path.LINETO]
        verts = [(pos_x, pos_y), (sub_x, sub_y)]
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, color=self._load_line_color, lw=self._load_line_width
        )
        self.ax.add_patch(patch)

    def _draw_load_bus(self, pos_x, pos_y, norm_dir_x, norm_dir_y, bus_id):
        center_x = pos_x + norm_dir_x * self._sub_radius
        center_y = pos_y + norm_dir_y * self._sub_radius
        face_color = self._line_bus_face_colors[bus_id]
        patch = patches.Circle(
            (center_x, center_y), radius=self._line_bus_radius, facecolor=face_color
        )
        self.ax.add_patch(patch)

    def draw_load(
        self,
        figure,
        observation,
        load_id,
        load_name,
        load_bus,
        load_value,
        load_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        self.xlim[0] = min(self.xlim[0], pos_x - self._load_radius)
        self.xlim[1] = max(self.xlim[1], pos_x + self._load_radius)
        self.ylim[0] = min(self.ylim[0], pos_y - self._load_radius)
        self.ylim[1] = max(self.ylim[1], pos_y + self._load_radius)
        self._draw_load_line(pos_x, pos_y, sub_x, sub_y)
        self._draw_load_circle(pos_x, pos_y)
        load_txt = ""
        if self._load_name:
            load_txt += '"{}":\n'.format(load_name)
        if self._load_id:
            load_txt += "id: {}\n".format(load_id)
        if load_value is not None:
            load_txt += pltu.format_value_unit(load_value, load_unit)
        if load_txt:
            self._draw_load_txt(pos_x, pos_y, sub_x, sub_y, load_txt)
        if self._display_load_name:
            self._draw_load_name(pos_x, pos_y, str(load_id))
        load_dir_x, load_dir_y = pltu.norm_from_points(sub_x, sub_y, pos_x, pos_y)
        self._draw_load_bus(sub_x, sub_y, load_dir_x, load_dir_y, load_bus)

    def update_load(
        self,
        figure,
        observation,
        load_id,
        load_name,
        load_bus,
        load_value,
        load_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        pass

    def draw_storage(
        self,
        figure,
        observation,
        load_id,
        load_name,
        load_bus,
        load_value,
        load_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        self.xlim[0] = min(self.xlim[0], pos_x - self._load_radius)
        self.xlim[1] = max(self.xlim[1], pos_x + self._load_radius)
        self.ylim[0] = min(self.ylim[0], pos_y - self._load_radius)
        self.ylim[1] = max(self.ylim[1], pos_y + self._load_radius)
        self._draw_storage_line(
            pos_x, pos_y, sub_x, sub_y
        )  # line from the storage to the substation
        self._draw_storage_circle(pos_x, pos_y)  # storage element

        load_txt = ""
        if self._storage_name:
            load_txt += '"{}":\n'.format(load_name)
        if self._storage_id:
            load_txt += "id: {}\n".format(load_id)
        if load_value is not None:
            load_txt += pltu.format_value_unit(load_value, load_unit)
        if load_txt:
            self._draw_load_txt(pos_x, pos_y, sub_x, sub_y, load_txt)
        if self._display_load_name:
            self._draw_load_name(pos_x, pos_y, str(load_id))
        load_dir_x, load_dir_y = pltu.norm_from_points(sub_x, sub_y, pos_x, pos_y)
        self._draw_storage_bus(sub_x, sub_y, load_dir_x, load_dir_y, load_bus)

    def _draw_storage_circle(self, pos_x, pos_y):
        patch = self._storage_patch(
            (pos_x, pos_y),
            radius=self._storage_radius,
            facecolor=self._storage_face_color,
            edgecolor=self._storage_edge_color,
        )
        self.ax.add_patch(patch)

    def _draw_storage_line(self, pos_x, pos_y, sub_x, sub_y):
        codes = [Path.MOVETO, Path.LINETO]
        verts = [(pos_x, pos_y), (sub_x, sub_y)]
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, color=self._storage_line_color, lw=self._storage_line_width
        )
        self.ax.add_patch(patch)

    def _draw_storage_bus(self, pos_x, pos_y, norm_dir_x, norm_dir_y, bus_id):
        center_x = pos_x + norm_dir_x * self._sub_radius
        center_y = pos_y + norm_dir_y * self._sub_radius
        face_color = self._line_bus_face_colors[bus_id]
        patch = patches.Circle(
            (center_x, center_y), radius=self._line_bus_radius, facecolor=face_color
        )
        self.ax.add_patch(patch)

    def update_storage(
        self,
        figure,
        observation,
        storage_name,
        storage_id,
        storage_bus,
        storage_value,
        storage_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        pass

    def _draw_gen_txt(self, pos_x, pos_y, sub_x, sub_y, text):
        dir_x, dir_y = pltu.vec_from_points(sub_x, sub_y, pos_x, pos_y)
        off_x, off_y = pltu.norm_from_points(sub_x, sub_y, pos_x, pos_y)
        txt_x = pos_x + off_x * self._gen_radius
        txt_y = pos_y + off_y * self._gen_radius
        ha = self._h_textpos_from_dir(dir_x, dir_y)
        va = self._v_textpos_from_dir(dir_x, dir_y)
        self.ax.text(
            txt_x,
            txt_y,
            text,
            color=self._gen_txt_color,
            wrap=True,
            fontsize="small",
            horizontalalignment=ha,
            verticalalignment=va,
        )

    def _draw_gen_circle(self, pos_x, pos_y, gen_edgecolor):
        patch = self._gen_patch(
            (pos_x, pos_y),
            radius=self._gen_radius,
            edgecolor=gen_edgecolor,
            facecolor=self._gen_face_color,
        )
        self.ax.add_patch(patch)

    def _draw_gen_line(self, pos_x, pos_y, sub_x, sub_y):
        codes = [Path.MOVETO, Path.LINETO]
        verts = [(pos_x, pos_y), (sub_x, sub_y)]
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, color=self._gen_line_color, lw=self._load_line_width
        )
        self.ax.add_patch(patch)

    def _draw_gen_name(self, pos_x, pos_y, txt):
        self.ax.text(
            pos_x,
            pos_y,
            txt,
            color=self._gen_txt_color,
            va="center",
            ha="center",
            fontsize="x-small",
        )

    def _draw_gen_bus(self, pos_x, pos_y, norm_dir_x, norm_dir_y, bus_id):
        center_x = pos_x + norm_dir_x * self._sub_radius
        center_y = pos_y + norm_dir_y * self._sub_radius
        face_color = self._line_bus_face_colors[bus_id]
        patch = patches.Circle(
            (center_x, center_y), radius=self._line_bus_radius, facecolor=face_color
        )
        self.ax.add_patch(patch)

    def draw_gen(
        self,
        figure,
        observation,
        gen_id,
        gen_name,
        gen_bus,
        gen_value,
        gen_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        self.xlim[0] = min(self.xlim[0], pos_x - self._gen_radius)
        self.xlim[1] = max(self.xlim[1], pos_x + self._gen_radius)
        self.ylim[0] = min(self.ylim[0], pos_y - self._gen_radius)
        self.ylim[1] = max(self.ylim[1], pos_y + self._gen_radius)
        hide = False
        if isinstance(self._gen_edge_color, str):
            # case where the color of the generator is a string (same color for all generators)
            gen_color = self._gen_edge_color
        else:
            my_val = observation.prod_p[gen_id]
            n_colors = len(self._gen_edge_color) - 1
            if np.isfinite(my_val):
                color_idx = max(0, min(n_colors, int(my_val * n_colors)))
            else:
                color_idx = 0
                hide = True
            gen_color = self._gen_edge_color[color_idx]

        if not hide:
            self._draw_gen_line(pos_x, pos_y, sub_x, sub_y)
            self._draw_gen_circle(pos_x, pos_y, gen_color)
            gen_txt = ""
            if self._gen_name:
                gen_txt += '"{}":\n'.format(gen_name)
            if self._gen_id:
                gen_txt += "id: {}\n".format(gen_id)
            if gen_value is not None and self._display_gen_value:
                gen_txt += pltu.format_value_unit(gen_value, gen_unit)
            if gen_txt:
                self._draw_gen_txt(pos_x, pos_y, sub_x, sub_y, gen_txt)
            if self._display_gen_name:
                self._draw_gen_name(pos_x, pos_y, str(gen_id))
            gen_dir_x, gen_dir_y = pltu.norm_from_points(sub_x, sub_y, pos_x, pos_y)
            self._draw_gen_bus(sub_x, sub_y, gen_dir_x, gen_dir_y, gen_bus)

    def update_gen(
        self,
        figure,
        observation,
        gen_id,
        gen_name,
        gen_bus,
        gen_value,
        gen_unit,
        pos_x,
        pos_y,
        sub_x,
        sub_y,
    ):
        pass

    def _draw_powerline_txt(self, pos_or_x, pos_or_y, pos_ex_x, pos_ex_y, text):
        pos_x, pos_y = pltu.middle_from_points(pos_or_x, pos_or_y, pos_ex_x, pos_ex_y)
        off_x, off_y = pltu.orth_norm_from_points(
            pos_or_x, pos_or_y, pos_ex_x, pos_ex_y
        )
        txt_x = pos_x + off_x * (self._load_radius / 2)
        txt_y = pos_y + off_y * (self._load_radius / 2)
        ha = self._h_textpos_from_dir(off_x, off_y)
        va = self._v_textpos_from_dir(off_x, off_y)
        self.ax.text(
            txt_x,
            txt_y,
            text,
            color=self._gen_txt_color,
            fontsize="small",
            horizontalalignment=ha,
            verticalalignment=va,
        )

    def _draw_powerline_line(
        self, pos_or_x, pos_or_y, pos_ex_x, pos_ex_y, color, line_style
    ):
        codes = [Path.MOVETO, Path.LINETO]
        verts = [(pos_or_x, pos_or_y), (pos_ex_x, pos_ex_y)]
        path = Path(verts, codes)
        patch = patches.PathPatch(
            path, color=color, lw=self._line_color_width, ls=line_style
        )
        self.ax.add_patch(patch)

    def _draw_powerline_bus(self, pos_x, pos_y, norm_dir_x, norm_dir_y, bus_id):
        center_x = pos_x + norm_dir_x * self._sub_radius
        center_y = pos_y + norm_dir_y * self._sub_radius
        face_color = self._line_bus_face_colors[bus_id]
        patch = patches.Circle(
            (center_x, center_y), radius=self._line_bus_radius, facecolor=face_color
        )
        self.ax.add_patch(patch)

    def _draw_powerline_arrow(
        self, pos_or_x, pos_or_y, pos_ex_x, pos_ex_y, color, watt_value
    ):
        sign = 1.0 if watt_value > 0.0 else -1.0
        off = 1.0 if watt_value > 0.0 else 2.0
        dx, dy = pltu.norm_from_points(pos_or_x, pos_or_y, pos_ex_x, pos_ex_y)
        lx = dx * self._line_arrow_len
        ly = dy * self._line_arrow_len
        arr_x = pos_or_x + dx * self._sub_radius + off * lx
        arr_y = pos_or_y + dy * self._sub_radius + off * ly
        patch = patches.FancyArrow(
            arr_x,
            arr_y,
            sign * lx,
            sign * ly,
            length_includes_head=True,
            head_length=self._line_arrow_len,
            head_width=self._line_arrow_width,
            edgecolor=color,
            facecolor=color,
        )
        self.ax.add_patch(patch)

    def assign_line_palette(
        self, palette_name="YlOrRd", nb_color=10, line_color_scheme=None
    ):
        """
        Assign a new color palette when you want to plot information on the powerline.

        Parameters
        ----------
        palette_name: ``str``
            Name of the Maplotlib.plyplot palette to use (name forwarded to `plt.get_cmap(palette_name)`)
        nb_color: ``int``
            Number of color to use

        Examples
        -------
        .. code-block:: python

            # color a grid based on the value of the thermal limit
            plot_helper.assign_line_palette(nb_color=100)

            # plot this grid
            _ = plot_helper.plot_info(line_values=env.get_thermal_limit(), line_unit="A", coloring="line")

            # restore the default coloring (["blue", "orange", "red"])
            plot_helper.restore_line_palette()

        Notes
        -----
        Some palette are available there `colormaps <https://matplotlib.org/tutorials/colors/colormaps.html>`_

        """
        if line_color_scheme is None:
            palette = plt.get_cmap(palette_name)
            cols = []
            for i in range(1, nb_color + 1):
                cols.append(palette(i / nb_color))
            self._line_color_scheme = cols
        else:
            self._line_color_scheme = copy.deepcopy(line_color_scheme)

    def restore_line_palette(self):
        self._line_color_scheme = self._line_color_scheme_orig

    def assign_gen_palette(
        self,
        palette_name="YlOrRd",
        nb_color=10,
        increase_gen_size=None,
        gen_line_width=None,
    ):
        """
        Assign a new color palette when you want to plot information on the generator.

        Parameters
        ----------
        palette_name: ``str``
            Name of the Maplotlib.plyplot palette to use (name forwarded to `plt.get_cmap(palette_name)`)
        nb_color: ``int``
            Number of color to use

        increase_gen_size: ``float``
            Whether or not to increase the generator sizes (``None`` to disable this feature, 1 has no effect)

        gen_line_width: ``float``
            Increase the width of the generator (if not ``None``)

        Examples
        -------
        .. code-block:: python

            # color a grid based on the value of the thermal limit
            plot_helper.assign_gen_palette(nb_color=100)

            # plot this grid
            _ = plot_helper.plot_info(gen_values=env.gen_pmax, coloring="gen")

            # restore the default coloring (all green)
            plot_helper.restore_gen_palette()

        Notes
        -----
        Some palette are available there `colormaps <https://matplotlib.org/tutorials/colors/colormaps.html>`_

        """
        if palette_name is not None and nb_color > 0:
            # the user changed the palette
            palette = plt.get_cmap(palette_name)
            cols = []
            for i in range(1, nb_color + 1):
                cols.append(palette(i / nb_color))
            self._gen_edge_color = cols
        if increase_gen_size is not None:
            # the user changed the generator sizes
            self._gen_radius = float(increase_gen_size) * self._gen_radius_orig
        if gen_line_width is not None:
            # the user changed the generator line width
            self._gen_line_width = float(gen_line_width)

    def restore_gen_palette(self):
        """restore every properties of the default generator layout"""
        self._gen_edge_color = self._gen_edge_color_orig
        self._gen_radius = self._gen_radius_orig
        self._gen_line_width = self._gen_line_width_orig

    def draw_powerline(
        self,
        figure,
        observation,
        line_id,
        line_name,
        connected,
        line_value,
        line_unit,
        or_bus,
        pos_or_x,
        pos_or_y,
        ex_bus,
        pos_ex_x,
        pos_ex_y,
    ):
        rho = observation.rho[line_id]
        n_colors = len(self._line_color_scheme) - 1
        hide = False
        if np.isfinite(rho):
            color_idx = max(0, min(n_colors, int(rho * n_colors)))
        else:
            color_idx = 0
            hide = True

        color = "black"
        if not hide:
            if connected and rho > 0.0:
                color = self._line_color_scheme[color_idx]
            line_style = "-" if connected else "--"
            self._draw_powerline_line(
                pos_or_x, pos_or_y, pos_ex_x, pos_ex_y, color, line_style
            )
            # Deal with line text configurations
            txt = ""
            if self._line_name:
                txt += '"{}"\n'.format(line_name)
            if self._line_id:
                txt += "id: {}\n".format(str(line_id))
            if line_value is not None:
                txt += pltu.format_value_unit(line_value, line_unit)
            if txt:
                self._draw_powerline_txt(pos_or_x, pos_or_y, pos_ex_x, pos_ex_y, txt)

            or_dir_x, or_dir_y = pltu.norm_from_points(
                pos_or_x, pos_or_y, pos_ex_x, pos_ex_y
            )
            self._draw_powerline_bus(pos_or_x, pos_or_y, or_dir_x, or_dir_y, or_bus)
            ex_dir_x, ex_dir_y = pltu.norm_from_points(
                pos_ex_x, pos_ex_y, pos_or_x, pos_or_y
            )
            self._draw_powerline_bus(pos_ex_x, pos_ex_y, ex_dir_x, ex_dir_y, ex_bus)
            watt_value = observation.p_or[line_id]
            if rho > 0.0 and watt_value != 0.0:
                self._draw_powerline_arrow(
                    pos_or_x, pos_or_y, pos_ex_x, pos_ex_y, color, watt_value
                )

    def update_powerline(
        self,
        figure,
        observation,
        line_id,
        line_name,
        connected,
        line_value,
        line_unit,
        or_bus,
        pos_or_x,
        pos_or_y,
        ex_bus,
        pos_ex_x,
        pos_ex_y,
    ):
        pass

    def _get_gen_legend(self):
        """super complex function to display the proper shape in the legend"""
        if isinstance(self._gen_edge_color, str):
            gen_legend_col = self._gen_edge_color
        else:
            gen_legend_col = self._gen_edge_color[int(len(self._gen_edge_color) / 2)]
        my_res = self._gen_resolution

        class GenObjectHandler:
            def legend_artist(self, legend, orig_handle, fontsize, handlebox):
                xdescent, ydescent = handlebox.xdescent, handlebox.ydescent
                width, height = handlebox.width, handlebox.height
                center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
                pp_ = GenDraw(
                    xy=center,
                    radius=min(width, height),
                    facecolor="w",
                    edgecolor=gen_legend_col,
                    transform=handlebox.get_transform(),
                    resolution=my_res,
                )
                handlebox.add_artist(pp_)
                return pp_

        gen_legend = self._gen_patch(
            (0, 0),
            facecolor=self._gen_face_color,
            edgecolor=gen_legend_col,
            radius=self._gen_radius,
        )
        return gen_legend, GenObjectHandler()

    def _get_load_legend(self):
        """super complex function to display the proper shape in the legend"""
        if isinstance(self._load_edge_color, str):
            load_legend_col = self._load_edge_color
        else:
            load_legend_col = self._load_edge_color[int(len(self._load_edge_color) / 2)]
        my_res = self._load_resolution

        class LoadObjectHandler:
            def legend_artist(self, legend, orig_handle, fontsize, handlebox):
                xdescent, ydescent = handlebox.xdescent, handlebox.ydescent
                width, height = handlebox.width, handlebox.height
                center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
                pp_ = LoadDraw(
                    xy=center,
                    radius=min(width, height),
                    facecolor="w",
                    edgecolor=load_legend_col,
                    transform=handlebox.get_transform(),
                    resolution=my_res,
                )
                handlebox.add_artist(pp_)
                return pp_

        load_legend = self._load_patch(
            (0, 0),
            facecolor=self._load_face_color,
            edgecolor=load_legend_col,
            radius=self._load_radius,
        )
        return load_legend, LoadObjectHandler()

    def _get_storage_legend(self):
        """super complex function to display the proper shape in the legend"""
        if isinstance(self._storage_edge_color, str):
            storage_legend_col = self._storage_edge_color
        else:
            storage_legend_col = self._storage_edge_color[
                int(len(self._storage_edge_color) / 2)
            ]
        my_res = self._storage_resolution

        class StorageObjectHandler:
            def legend_artist(self, legend, orig_handle, fontsize, handlebox):
                xdescent, ydescent = handlebox.xdescent, handlebox.ydescent
                width, height = handlebox.width, handlebox.height
                center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
                pp_ = StorageDraw(
                    xy=center,
                    radius=min(width, height),
                    facecolor="w",
                    edgecolor=storage_legend_col,
                    transform=handlebox.get_transform(),
                    resolution=my_res,
                )
                handlebox.add_artist(pp_)
                return pp_

        storage_legend = self._storage_patch(
            (0, 0),
            facecolor=self._storage_face_color,
            edgecolor=storage_legend_col,
            radius=self._storage_radius,
        )
        return storage_legend, StorageObjectHandler()

    def draw_legend(self, figure, observation):
        title_str = observation.env_name
        if hasattr(observation, "month"):
            title_str = "{:02d}/{:02d} {:02d}:{:02d}".format(
                observation.day,
                observation.month,
                observation.hour_of_day,
                observation.minute_of_hour,
            )

        # generate the right legend for generator
        gen_legend, gen_handler = self._get_gen_legend()
        # generate the correct legend for load
        load_legend, load_handler = self._get_load_legend()
        # generate the correct legend for storage
        storage_legend, storage_handler = self._get_storage_legend()

        legend_help = [
            Line2D([0], [0], color="black", lw=1),
            Line2D([0], [0], color=self._sub_edge_color, lw=3),
            load_legend,
            gen_legend,
            storage_legend,
            Line2D([0], [0], marker="o", color=self._line_bus_face_colors[0]),
            Line2D([0], [0], marker="o", color=self._line_bus_face_colors[1]),
            Line2D([0], [0], marker="o", color=self._line_bus_face_colors[2]),
        ]
        self.legend = self.ax.legend(
            legend_help,
            [
                "powerline",
                "substation",
                "load",
                "generator",
                "storage",
                "no bus",
                "bus 1",
                "bus 2",
            ],
            title=title_str,
            handler_map={
                GenDraw: gen_handler,
                LoadDraw: load_handler,
                StorageDraw: storage_handler,
            },
        )
        # Hide axis
        self.ax.get_xaxis().set_visible(False)
        self.ax.get_yaxis().set_visible(False)
        # Hide frame
        self.ax.set(frame_on=False)

        # save the figure
        self.figure = figure

    def plot_postprocess(self, figure, observation, update):
        if not update:
            xmin = self.xlim[0] - self.xpad
            xmax = self.xlim[1] + self.xpad
            self.ax.set_xlim(xmin, xmax)
            ymin = self.ylim[0] - self.ypad
            ymax = self.ylim[1] + self.ypad
            self.ax.set_ylim(ymin, ymax)
            figure.tight_layout()

    def _save_plot_charact(self):
        _gen_edge_color_orig = self._gen_edge_color
        _gen_radius_orig = self._gen_radius
        _gen_line_width_orig = self._gen_line_width
        _display_gen_value = self._display_gen_value
        _display_gen_name = self._display_gen_name
        _display_sub_name = self._display_sub_name
        _display_load_name = self._display_load_name

        return (
            _gen_edge_color_orig,
            _gen_radius_orig,
            _gen_line_width_orig,
            _display_gen_value,
            _display_gen_name,
            _display_sub_name,
            _display_load_name,
        )

    def _restore_plot_charact(self, data):
        (
            _gen_edge_color_orig,
            _gen_radius_orig,
            _gen_line_width_orig,
            _display_gen_value,
            _display_gen_name,
            _display_sub_name,
            _display_load_name,
        ) = data
        self._gen_edge_color = _gen_edge_color_orig
        self._gen_radius = _gen_radius_orig
        self._gen_line_width = _gen_line_width_orig
        self._display_gen_value = _display_gen_value
        self._display_gen_name = _display_gen_name
        self._display_sub_name = _display_sub_name
        self._display_load_name = _display_load_name

    def plot_gen_type(self, increase_gen_size=1.5, gen_line_width=3):
        # save the sate of the generators config
        data = self._save_plot_charact()

        # do the plot
        self._display_gen_value = False
        self._display_gen_name = False
        self._display_sub_name = False
        self._display_load_name = False
        self.assign_gen_palette(
            nb_color=0,
            increase_gen_size=increase_gen_size,
            gen_line_width=gen_line_width,
        )
        self._gen_edge_color = [COLOR_GEN[i] for i in range(len(TYPE_GEN))]
        gen_values = [TYPE_GEN[el] for el in self.observation_space.gen_type]
        self.figure = self.plot_info(gen_values=gen_values, coloring="gen")
        self.add_legend_gentype()

        # restore the state to its initial configuration
        self._restore_plot_charact(data)

        return self.figure

    def plot_current_dispatch(
        self,
        obs,
        do_plot_actual_dispatch=True,
        increase_gen_size=1.5,
        gen_line_width=3,
        palette_name="coolwarm",
    ):
        # save the sate of the generators config
        data = self._save_plot_charact()

        # do the plot
        self._display_sub_name = False
        self._display_load_name = False
        self.assign_gen_palette(
            nb_color=5,
            palette_name=palette_name,
            increase_gen_size=increase_gen_size,
            gen_line_width=gen_line_width,
        )
        if do_plot_actual_dispatch:
            gen_values = obs.actual_dispatch
        else:
            gen_values = obs.target_dispatch
        self.figure = self.plot_info(
            gen_values=gen_values, coloring="gen", gen_unit="MW"
        )

        # restore the state to its initial configuration
        self._restore_plot_charact(data)

        return self.figure

    def add_legend_gentype(self, loc="lower right"):
        """add the legend for each generator type"""
        keys = sorted(TYPE_GEN.keys())
        ax_ = self.figure.axes[0]
        legend_help = [
            Line2D([0], [0], color=COLOR_GEN[TYPE_GEN[k]], label=k) for k in keys
        ]
        _ = ax_.legend(legend_help, keys, title="generator types", loc=loc)
        ax_.add_artist(self.legend)
